import graph_tool.all as gt
from collections import OrderedDict

from loguru import logger

from archx.interface import query_interface
from archx.utils import read_yaml, write_yaml, get_path, create_dir, get_prod


key_value = 'value'
key_unit = 'unit'
legal_aggregation = ['module', 'summation', 'specified']
legal_aggregation_tag = ['module', 'summation']
key_metric = 'metric'
key_instance = 'instance'
key_factor = 'factor'
key_aggregation = 'aggregation'
single_op_metric_format = '{\'' + key_value + '\': ' + 'float' + ', \'' + key_unit + '\': ' + 'str' + '}'


def create_metric_dict(metric_file: str) -> OrderedDict:
    """
    All metrics need to be include in this file
    """
    metric_file_full_path = get_path(metric_file)
    metric_dict = read_yaml(get_path(metric_file))[key_metric]

    for metric in metric_dict:
        # each metric requires unit, aggregation
        assert key_unit in metric_dict[metric], logger.error(f'Missing <unit> in metric <{metric}>.')
        
        if key_aggregation not in metric_dict[metric]:
            metric_dict[metric][key_aggregation] = 'summation'

        assert key_aggregation in metric_dict[metric], logger.error(f'Missing <aggregation> in metric <{metric}>.')
        this_aggregation = metric_dict[metric][key_aggregation].lower()
        assert this_aggregation in legal_aggregation, logger.error(f'Invalid aggregation <{this_aggregation}> in metric <{metric}>; legal values: {legal_aggregation}.')
        metric_dict[metric][key_aggregation] = this_aggregation
    
    logger.success(f'Create metric dictionary from <{metric_file_full_path}>.')

    return metric_dict


def save_metric_dict(metric_dict: OrderedDict, save_path: str) -> None:
    save_path = get_path(save_path, check_exist=False)
    metric_dict_ckpt= OrderedDict({key_metric: metric_dict})
    write_yaml(save_path, metric_dict_ckpt)
    logger.success(f'Save metric dictionary to <{save_path}>.')


def load_metric_dict(ckpt_path: str) -> OrderedDict:
    """
    Load metric from a checkpoint
    """
    full_path = get_path(ckpt_path)
    metric_dict = read_yaml(full_path)[key_metric]
    logger.success(f'Load metric dictionary from <{full_path}>.')
    return metric_dict


def create_event_metrics(event_graph: gt.Graph, architecture_dict: OrderedDict, metric_dict: OrderedDict, run_dir: str=None) -> gt.Graph:
    """
    Update the event graph, add metrics to each node.
    """

    # add event metric to vertex properties
    metric_property = event_graph.new_vertex_property('object')
    event_graph.vp.metric = metric_property

    # add tag to vertex properties
    tag_property = event_graph.new_vertex_property('object')
    event_graph.vp.tag = tag_property

    # initialize all metrics to 0.0
    for v in event_graph.vertices():
        event_graph.vp.metric[v] = OrderedDict({})
        for metric in metric_dict:
            event_graph.vp.metric[v][metric] = OrderedDict({key_value: 0.0, key_unit: metric_dict[metric][key_unit]})
        event_name = event_graph.vp.event[v]
        logger.info(f'Create metrics for event <{event_name}>.')

    logger.success(f'Create metrics for all events.')

    event_graph = create_module_metrics(event_graph, architecture_dict, run_dir)

    return event_graph


def create_module_metrics(event_graph: gt.Graph, architecture_dict: OrderedDict, run_dir: str=None) -> gt.Graph:
    """
    This function queries the interface for each architecture modules in the event graph
    """

    create_dir(run_dir)
    full_path = get_path(run_dir)

    # all modules have out degree of 0
    for v in gt.find_vertex(event_graph, 'out', 0):
        module_name = event_graph.vp.event[v]
        assert module_name in architecture_dict, logger.error(f'Invalid module <{module_name}>.')
        assert 'query' in architecture_dict[module_name], logger.error(f'Missing query information for module <{module_name}>.')
        assert 'class' in architecture_dict[module_name]['query'], logger.error(f'Missing class information in query for module <{module_name}>.')
        
        module_class = architecture_dict[module_name]['query']['class']
        query = architecture_dict[module_name]['query']
        result = query_interface(module_name, query, output_dir=full_path)

        # if query generates new results, update metric with new results
        event_graph.vp.metric[v] = OrderedDict({k: result.get(k, event_graph.vp.metric[v][k]) for k in event_graph.vp.metric[v]})
        
        # get number of instances for an architecture module
        event_graph.vp.metric[v][key_instance] = get_prod(architecture_dict[module_name][key_instance])

        event_graph.vp.tag[v] = architecture_dict[module_name]['tag']

        logger.info(f'Create metrics for module <{module_name}> with class <{module_class}>.')
    
    logger.success(f'Create metrics for all modules.')
    
    return event_graph


def query_module_metric(event_graph: gt.Graph, metric_dict: str, metric: str, module: str=None, operation: str=None) -> OrderedDict:
    """
    Query the metric of a module or an operation in the module.
    """
    assert metric in metric_dict, logger.error(f'Invalid metric <{metric}>.')
    assert key_aggregation in metric_dict[metric], logger.error(f'Missing aggregation in metric <{metric}>.')
    assert metric_dict[metric][key_aggregation] in legal_aggregation, logger.error(f'Invalid aggregation <{metric_dict[metric][key_aggregation]}> for metric <{metric}>; legal values: {legal_aggregation}.')
    
    # get the module node
    module_node = gt.find_vertex(event_graph, event_graph.vp.event, module)
    assert len(module_node) == 1, logger.error(f'Invalid module <{module}>.')
    module_node = module_node[0]
    assert module_node.out_degree() == 0, logger.error(f'Invalid event <{module}>; this function requires a module.')

    module_metric = event_graph.vp.metric[module_node][metric]

    # if a module has multiple operations, check if the operation is legal
    # this only works for certain modules, e.g., sram has multiple operations (read, write)
    legal_ops = []

    if (key_value not in module_metric) & (key_unit not in module_metric):
        for key in module_metric:
            legal_ops.append(key)

    
    if len(legal_ops) == 0:
        # single-operation module
        if operation is not None:
            logger.warning(f'Ignore operation <{operation}> for metric <{metric}> in module <{module}>; this module requires no specified operation.')
        query_result = module_metric
    else:
        # multi-operation module
        # example event metric:
        # OrderedDict([('read', OrderedDict([('value', 0.474466), ('unit', 'nJ')])), ('write', OrderedDict([('value', 0.499877), ('unit', 'nJ')]))])
        assert operation in legal_ops, logger.error(f'Invalid operation <{operation}> for metric <{metric}> in module <{module}>; legal values: {legal_ops}.')
        query_result = module_metric[operation]
    
    if operation is None:
        logger.success(f'Query metric <{metric}> for module <{module}>.')
    else:
        logger.success(f'Query metric <{metric}> for operation <{operation}> in module <{module}>.')
    
    return query_result


def aggregate_event_count(event_graph: gt.Graph, workload: str=None, event: str=None) -> OrderedDict:
    # find the number of event underworkload

    # get the event node
    event_node = gt.find_vertex(event_graph, event_graph.vp.event, event)
    assert len(event_node) > 0, logger.error(f'Invalid event <{event}>.')
    event_node = event_node[0]

    # validate event in workload
    if workload is not None and workload != event:
        # get the workload node
        workload_node = gt.find_vertex(event_graph, event_graph.vp.event, workload)
        assert len(workload_node) > 0, logger.error(f'Invalid workload <{workload}>.')
        workload_node = workload_node[0]
        paths = []
        for path in gt.all_paths(event_graph, workload_node, event_node):
            paths.append(path)
        assert len(paths) > 0, logger.error(f'Invalid event <{event}> in workload <{workload}>.')

        # get the event count from workload to event
        total_event_count = 0.
        for path in gt.all_paths(event_graph, workload_node, event_node):
            path_event_count = 1.
            # calculate total event count based on the path
            for idx in range(len(path)-1):
                edge = event_graph.edge(path[idx], path[idx+1])
                path_event_count *= event_graph.ep.count[edge]
                logger.debug(f'  Path (<{event_graph.vp.event[edge.source()]}> -> <{event_graph.vp.event[edge.target()]}>) updates event count to <{path_event_count}>.')
            total_event_count += path_event_count
            logger.debug(f'  Total event count (<{workload}> -> <{event}>) is increased by <{path_event_count}>.')
        # get the metric value from event to all its subevents
        logger.debug(f'  Total value (<{workload}> -> <{event}>) = <{total_event_count}>.')

    else:
        total_event_count = 1.
        logger.debug(f'  Total value (<{event}>) = <{total_event_count}>.')

    if workload is not None:
        logger.success(f'Aggregate event count for event <{event}> in workload <{workload}>.')
    else:
        logger.success(f'Aggregate event count for event <{event}>.')

    return total_event_count


def aggregate_event_metric(event_graph: gt.Graph, metric_dict: str, metric: str, workload: str=None, event: str=None) -> OrderedDict:
    """
    If workload is none, the metric will be aggregated to the event.
    If workload is not none, the metric will be aggregated to the workload.
    """
    assert metric in metric_dict, logger.error(f'Invalid metric <{metric}>.')
    assert key_aggregation in metric_dict[metric], logger.error(f'Missing aggregation in metric <{metric}>.')
    assert metric_dict[metric][key_aggregation] in legal_aggregation, logger.error(f'Invalid aggregation <{metric_dict[metric][key_aggregation]}> for metric <{metric}>; legal values: {legal_aggregation}.')
    
    metric_cfg = metric_dict[metric]
    
    # get the event node
    event_node = gt.find_vertex(event_graph, event_graph.vp.event, event)
    assert len(event_node) > 0, logger.error(f'Invalid event <{event}>.')
    event_node = event_node[0]

    # validate event in workload
    if workload is not None and workload != event:
        # get the workload node
        workload_node = gt.find_vertex(event_graph, event_graph.vp.event, workload)
        assert len(workload_node) > 0, logger.error(f'Invalid workload <{workload}>.')
        workload_node = workload_node[0]
        paths = []
        for path in gt.all_paths(event_graph, workload_node, event_node):
            paths.append(path)
        assert len(paths) > 0, logger.error(f'Invalid event <{event}> in workload <{workload}>.')

    # reset the metric values of all events to 0.0
    for v in event_graph.vertices():
        # module metric is not reset
        if v.out_degree() != 0:
            if event_graph.vp.metric[v][metric][key_value] != 0:
                logger.debug(f'  Reset metric <{metric}> in event <{event_graph.vp.event[v]}> to 0.')
            event_graph.vp.metric[v][metric][key_value] = 0.

    # output metric
    event_metric = OrderedDict({key_value: 0.0, key_unit: metric_cfg[key_unit]})

    aggregation = metric_cfg[key_aggregation]

    # module aggregation mode: only sum all leaf nodes from current node
    if aggregation == 'module':
        # the input can be either a module or an event, i.e., no limitations on the output degree of event_node
        # factor has no impact on aggregation in the module mode
        if workload is not None:
            logger.warning(f'Ignore workload <{workload}> in aggregation <{aggregation}>.')

        module_sum = 0.
        evaluated = []

        module_sum += aggregate_module(graph=event_graph, current_node=event_node, sum=0, metric=metric, evaluated=evaluated, top_event=event)
        
        event_metric[key_value] = module_sum
        if workload is None:
            logger.debug(f'  Total value (<{event}>) = <{event_metric[key_value]}> <{event_metric[key_unit]}>.')
        else:
            logger.debug(f'  Total value (<{workload}> -> <{event}>) = <{event_metric[key_value]}> <{event_metric[key_unit]}>.')

    # summation aggregation mode: sum all child nodes from current node
    elif aggregation == 'summation':
        # check if the event is a single-operation module
        is_single_operation, legal_ops = check_single_operation(event_graph=event_graph, metric=metric, event_node=event_node)

        if workload is None or workload == event:
            # if event is a module, summation aggregation requires single-operation module, since no operation is specified in multi-operation module
            if event_node.out_degree() == 0:
                assert is_single_operation, logger.error(f'Invalid module <{event}> for aggregation <{aggregation}>; this aggregation does not support multi-operation module.')
            
            event_graph = aggregate_summation(event_graph=event_graph, start_node=event_node, metric=metric)
            event_metric[key_value] = event_graph.vp.metric[event_node][metric][key_value]
            if workload is None:
                logger.debug(f'  Total value (<{event}>) = <{event_metric[key_value]}> <{event_metric[key_unit]}>.')
            else:
                logger.debug(f'  Total value (<{workload}> -> <{event}>) = <{event_metric[key_value]}> <{event_metric[key_unit]}>.')

        else:
            is_multi_operation_module = (not is_single_operation) & (event_node.out_degree() == 0)

            if is_multi_operation_module:
                legal_op_count_dict = OrderedDict()
                legal_op_metric_dict = OrderedDict()
                for legal_op in legal_ops:
                    legal_op_count_dict[legal_op] = 0.
                    legal_op_metric_dict[legal_op] = 0.

                for path in gt.all_paths(event_graph, workload_node, event_node):
                    path_event_count = 1.
                    path_event_factor = 1.
                    # calculate total event count based on the path
                    for idx in range(len(path)-1):
                        edge = event_graph.edge(path[idx], path[idx+1])
                        path_event_count *= event_graph.ep.count[edge]
                        if metric in event_graph.ep.factor[edge]:
                            path_event_factor *= event_graph.ep.factor[edge][metric]
                        logger.debug(f'  Path (<{event_graph.vp.event[edge.source()]}> -> <{event_graph.vp.event[edge.target()]}>) updates event count to <{path_event_count}> and event factor to <{path_event_factor}>.')
                    operation = event_graph.ep.operation[edge].get(metric).lower()
                    legal_op_metric_dict[operation] = get_metric_value(event_graph=event_graph, edge=edge, metric=metric)
                    delta_event_count = path_event_count * path_event_factor
                    legal_op_count_dict[operation] += delta_event_count
                    logger.debug(f'  Total event count (<{event}> : <{operation}>) is increased by <{delta_event_count}> = count <{path_event_count}> * factor <{path_event_factor}>.')

                for legal_op in legal_ops:
                    op_metric_value = legal_op_metric_dict[legal_op] * legal_op_count_dict[legal_op]
                    event_metric[key_value] += op_metric_value
                    logger.debug(f'  Total value (<{event}> : <{legal_op}>) = <{op_metric_value}> <{event_metric[key_unit]}> = single value <{legal_op_metric_dict[legal_op]}> * count <{legal_op_count_dict[legal_op]}>.')
                logger.debug(f'  Total value (<{workload}> -> <{event}>) = <{event_metric[key_value]}> <{event_metric[key_unit]}>.')

            else:
                # get the event count from workload to event
                total_event_count = 0.
                for path in gt.all_paths(event_graph, workload_node, event_node):
                    path_event_count = 1.
                    path_event_factor = 1.
                    # calculate total event count based on the path
                    for idx in range(len(path)-1):
                        edge = event_graph.edge(path[idx], path[idx+1])
                        path_event_count *= event_graph.ep.count[edge]
                        if metric in event_graph.ep.factor[edge]:
                            path_event_factor *= event_graph.ep.factor[edge][metric]
                        logger.debug(f'  Path (<{event_graph.vp.event[edge.source()]}> -> <{event_graph.vp.event[edge.target()]}>) updates event count to <{path_event_count}> and event factor to <{path_event_factor}>.')
                    delta_event_count = path_event_count * path_event_factor
                    total_event_count += delta_event_count
                    logger.debug(f'  Total event count (<{workload}> -> <{event}>) is increased by <{delta_event_count}> = count <{path_event_count}> * factor <{path_event_factor}>.')
                # get the metric value from event to all its subevents
                event_graph = aggregate_summation(event_graph=event_graph, start_node=event_node, metric=metric)
                event_metric[key_value] = event_graph.vp.metric[event_node][metric][key_value] * total_event_count
                logger.debug(f'  Total value (<{workload}> -> <{event}>) = <{event_metric[key_value]}> <{event_metric[key_unit]}> = single value <{event_graph.vp.metric[event_node][metric][key_value]}> * count <{total_event_count}>.')

    # specified aggregation mode: parallel/sequential is taken into account
    elif aggregation == 'specified':
        # for specified mode, the input can not be a module, i.e., the output degree shall be large than 0
        assert event_node.out_degree() > 0, logger.error(f'Invalid module <{event}> for aggregation <{aggregation}>; this aggregation requires an event.')
        
        if workload is None or workload == event:
            total_event_count = 1.
        else:
            # get the event count from workload to event
            total_event_count = 0.
            for path in gt.all_paths(event_graph, workload_node, event_node):
                path_event_count = 1.
                path_event_factor = 1.
                # calculate total event count based on the path
                for idx in range(len(path)-1):
                    edge = event_graph.edge(path[idx], path[idx+1])
                    path_event_count *= event_graph.ep.count[edge]
                    if metric in event_graph.ep.factor[edge]:
                        path_event_factor *= event_graph.ep.factor[edge][metric]
                    logger.debug(f'  Path (<{event_graph.vp.event[edge.source()]}> -> <{event_graph.vp.event[edge.target()]}>) updates event count to <{path_event_count}> and event factor to <{path_event_factor}>.')
                delta_event_count = path_event_count * path_event_factor
                total_event_count += delta_event_count
                logger.debug(f'  Total event count (<{workload}> -> <{event}>) is increased by <{delta_event_count}> = count <{path_event_count}> * factor <{path_event_factor}>.')
        
        event_graph = aggregate_specified(event_graph=event_graph, start_node=event_node, metric=metric)
        event_metric[key_value] = event_graph.vp.metric[event_node][metric][key_value] * total_event_count
        if workload is None:
            logger.debug(f'  Total value (<{event}>) = <{event_metric[key_value]}> <{event_metric[key_unit]}> = single value <{event_graph.vp.metric[event_node][metric][key_value]}>.')
        else:
            logger.debug(f'  Total value (<{workload}> -> <{event}>) = <{event_metric[key_value]}> <{event_metric[key_unit]}> = single value <{event_graph.vp.metric[event_node][metric][key_value]}> * count <{total_event_count}>.')
    
    if workload is not None:
        logger.success(f'Aggregate metric <{metric}> for event <{event}> in workload <{workload}> with aggregation <{aggregation}>.')
    else:
        logger.success(f'Aggregate metric <{metric}> for event <{event}> with aggregation <{aggregation}>.')

    return event_metric


def aggregate_module(graph: gt.Graph, current_node: gt.Vertex, sum: float, metric: str, evaluated: list, top_event: str) -> float:
    # leaf nodes will return their own value
    if current_node.out_degree() == 0:
        if current_node not in evaluated:
            key_list = list(graph.vp.metric[current_node][metric].keys())
            key_list.sort()
            expected_key_list = [key_value, key_unit]
            expected_key_list.sort()
            assert key_list == expected_key_list, logger.error(f'Invalid metric <{metric}> for module <{graph.vp.event[current_node]}>.')
            evaluated.append(current_node)
            logger.info(f'Aggregate metric <{metric}> for module <{graph.vp.event[current_node]}> in event <{top_event}>.')
            total_metric_value = graph.vp.metric[current_node][metric][key_value] * graph.vp.metric[current_node][key_instance]
            logger.debug(f'  Total value (<{top_event}> -> <{graph.vp.event[current_node]}>) = <{total_metric_value}> <{graph.vp.metric[current_node][metric][key_unit]}> = single value <{graph.vp.metric[current_node][metric][key_value]}> * instance <{graph.vp.metric[current_node][key_instance]}>.')
            return sum + total_metric_value
        else:
            logger.info(f'Ignore repeated module <{graph.vp.event[current_node]}>.')
            return 0
    
    # non-leaf nodes will aggregate their children
    else:
        temp_sum = 0.
        for child in current_node.out_neighbors():
            logger.debug(f'  Check subevent <{graph.vp.event[current_node]}> -> <{graph.vp.event[child]}>.')
            temp_sum += aggregate_module(graph, child, sum, metric, evaluated, graph.vp.event[current_node])
        return temp_sum


def aggregate_summation(event_graph: gt.Graph, start_node: gt.Vertex, metric: str) -> gt.Graph:
    # topological sort for aggregation, from module to start event
    topo_order = topological_sort_reverse(event_graph, start_node=start_node)

    # aggregate metric upwards through the graph
    for v in topo_order:
        if event_graph.vertex(v).out_degree() == 0:
            logger.info(f'Aggregate metric <{metric}> for module <{event_graph.vp.event[v]}>.')
        else:
            logger.info(f'Aggregate metric <{metric}> for event <{event_graph.vp.event[v]}>.')

        for e in event_graph.vertex(v).in_edges():
            # only calculate if the parent event is on the path (in topo_order)
            if e.source() in topo_order:
                # the count of subevents in event
                subevent_count = event_graph.ep.count[e]
                if metric in event_graph.ep.factor[e]:
                    event_factor = event_graph.ep.factor[e][metric]
                else:
                    event_factor = 1.
                # get the metric value of target node
                edge_target_metric = get_metric_value(event_graph=event_graph, edge=e, metric=metric)
                # update the metric value of source node
                total_metric_value = subevent_count * edge_target_metric * event_factor
                event_graph.vp.metric[e.source()][metric][key_value] += total_metric_value
                logger.debug(f'  Total value (<{event_graph.vp.event[e.source()]}> -> <{event_graph.vp.event[e.target()]}>) = <{total_metric_value}> <{event_graph.vp.metric[e.source()][metric][key_unit]}> = single value <{edge_target_metric}> * count <{subevent_count}> * factor <{event_factor}>.')

    return event_graph


def aggregate_specified(event_graph: gt.Graph, start_node: gt.Vertex, metric: str) -> gt.Graph:
    # topological sort for aggregation, from module to start event
    # start from the event connected to modules and ignore modules
    topo_order = topological_sort_reverse(event_graph, start_node=start_node)

    # aggregate metric upwards through the graph
    for v in topo_order:
        event_name = event_graph.vp.event[v]
        if event_graph.vertex(v).out_degree() == 0:
            logger.info(f'Ignore module <{event_name}>.')
        else:
            logger.info(f'Aggregate metric <{metric}> for event <{event_name}>.')

            parallel_max = 0.
            sequential_acc = 0.

            connect_leaf_only = True
            connect_leaf_any = False

            for e in event_graph.vertex(v).out_edges():
                edge_target_event = event_graph.vp.event[e.target()]
                if e.target().out_degree() == 0:
                    assert key_value in event_graph.vp.metric[e.target()][metric], logger.error(f'Invalid metric <{metric}> for event <{edge_target_event}>; legal metric: {single_op_metric_format}.')
                    connect_leaf_any = True
                    logger.debug(f'  Ignore module <{edge_target_event}>.')
                
                else:
                    connect_leaf_only = False
                    metric_mode = event_graph.ep.aggregation[e]
                    if metric in event_graph.ep.factor[e]:
                        event_factor = event_graph.ep.factor[e][metric]
                    else:
                        event_factor = 1.
                    metric_value = event_graph.vp.metric[e.target()][metric][key_value] * event_graph.ep.count[e] * event_factor
                    logger.debug(f'  Total value (<{edge_target_event}>) = <{metric_value}> <{event_graph.vp.metric[e.target()][metric][key_unit]}> = single value <{event_graph.vp.metric[e.target()][metric][key_value]}> * count <{event_graph.ep.count[e]}> * factor <{event_factor}>.')

                    if metric_mode == 'parallel':
                        if metric_value > parallel_max:
                            logger.debug(f'  Update parallel maximum metric value to <{metric_value}>, based on event <{edge_target_event}>.')
                            parallel_max = metric_value
                    else:
                        logger.debug(f'  Update sequential accumulated metric value by <{metric_value}>, based on event <{edge_target_event}>.')
                        sequential_acc += metric_value
            
            # the final value is the sum of sequential acc and maximum parallel
            event_graph.vp.metric[v][metric][key_value] = sequential_acc + parallel_max

            # report design errors in the performance model
            # case1: if an event is only connected to modules, it should have a performance model with metric defined
            if connect_leaf_only is True:
                assert event_graph.vp[metric][v] is not None, logger.error(f'  Missing metric <{metric}> in event <{event_graph.vp.event[v]}>, since it is only connected to modules; check the performance model.')
                event_graph.vp.metric[v][metric][key_value] = event_graph.vp[metric][v][key_value]
            
            # case2: if an event is connected to no modules, it should not have a performance model with metric defined
            if connect_leaf_any is False:
                assert event_graph.vp[metric][v] is None, logger.error(f'  Invalid metric <{metric}> in event <{event_graph.vp.event[v]}>, since it is connected to no modules; check the performance model.')
            
            logger.debug(f'  Total value (<{event_graph.vp.event[v]}>) = <{event_graph.vp.metric[v][metric][key_value]}> <{event_graph.vp.metric[v][metric][key_unit]}>.')

    return event_graph


def topological_sort_reverse(graph: gt.Graph, start_node: gt.Vertex) -> list:
    # topological sort with dfs
    # output is not deterministic due to use of dfs
    topo_order = list(gt.topological_sort(graph))

    # remove redundant nodes from the start event
    to_remove = []
    for v in topo_order:
        if v != start_node:
            node_list = gt.shortest_path(graph, start_node, v)[0]
            if len(node_list) == 0:
                to_remove.append(v)
    for v in to_remove:
        topo_order.remove(v)
    
    topo_order.reverse()

    return topo_order


def get_metric_value(event_graph: gt.Graph, edge: gt.Edge, metric: str) -> float:
    target_event_node = edge.target()
    edge_target_event = event_graph.vp.event[target_event_node]
    event_metric = event_graph.vp.metric[target_event_node][metric]

    is_single_operation, legal_ops = check_single_operation(event_graph=event_graph, metric=metric, event_node=target_event_node)
    
    if is_single_operation:
        # event or single-operation module
        return event_metric[key_value]
    else:
        # multi-operation module
        # example event metric:
        # OrderedDict([('read', OrderedDict([('value', 0.474466), ('unit', 'nJ')])), ('write', OrderedDict([('value', 0.499877), ('unit', 'nJ')]))])
        operation = event_graph.ep.operation[edge].get(metric).lower()
        assert operation in legal_ops, logger.error(f'Invalid operation <{operation}> for metric <{metric}> in module <{edge_target_event}>; legal values: {legal_ops}.')
        return event_metric[operation][key_value]
    

def check_single_operation(event_graph: gt.Graph, metric: str, event_node: gt.Edge) -> bool:
    is_single_operation = False
    legal_ops = []

    if (key_value not in event_graph.vp.metric[event_node][metric]) | (key_unit not in event_graph.vp.metric[event_node][metric]):
        for key in event_graph.vp.metric[event_node][metric]:
            legal_ops.append(key)
    else:
        assert len(event_graph.vp.metric[event_node][metric]) == 2, logger.error(f'Invalid metric <{metric}> for event <{event_graph.vp.event[event_node]}>; legal metric: {single_op_metric_format}.')

    if len(legal_ops) == 0:
        is_single_operation = True
    return is_single_operation, legal_ops


def aggregate_tag_metric(event_graph: gt.Graph, metric_dict: str, metric: str, workload: str=None, tag: str=None) -> OrderedDict:
    """
    Aggregate the metric according to a tag.
    if workload is none, the metric will be aggregated to the tag.
    if workload is not none, the metric will be aggregated to the workload.
    """
    
    assert tag is not None, logger.error(f'Missing tag in aggregation of metric <{metric}>.')
    assert metric in metric_dict, logger.error(f'Invalid metric <{metric}>.')
    assert key_aggregation in metric_dict[metric], logger.error(f'Missing aggregation in metric <{metric}>.')
    assert metric_dict[metric][key_aggregation] != 'specified', logger.error(f'Invalid aggregation <{metric_dict[metric][key_aggregation]}> for tag <{tag}>; legal values: {legal_aggregation_tag}.')

    # find all modules with the tag
    tag_nodes = []
    for v in gt.find_vertex(event_graph, 'out', 0):
        if tag in event_graph.vp.tag[v]:
            tag_nodes.append(v)
    assert len(tag_nodes) > 0, logger.error(f'Invalid tag <{tag}>.')

    tag_metric = OrderedDict({key_value: 0.0, key_unit: metric_dict[metric][key_unit]})

    # for each module with the tag, aggregate the metric
    for tag_node in tag_nodes:
        module_name = event_graph.vp.event[tag_node]
        module_metric = aggregate_event_metric(event_graph=event_graph, metric_dict=metric_dict, metric=metric, workload=workload, event=module_name)
        assert tag_metric[key_unit] == module_metric[key_unit], logger.error(f'Inconsistent unit in metric <{metric}> for module <{module_name}> with tag <{tag}>.')
        tag_metric[key_value] += module_metric[key_value]
        logger.debug(f'  Total value (module <{module_name}>) = <{module_metric[key_value]}> <{module_metric[key_unit]}>.')
    
    logger.debug(f'  Total value (tag <{tag}>) = <{tag_metric[key_value]}> <{tag_metric[key_unit]}>.')

    if workload is None:
        logger.success(f'Aggregate metric <{metric}> for tag <{tag}>.')
    else:
        logger.success(f'Aggregate metric <{metric}> for tag <{tag}> in workload <{workload}>.')

    return tag_metric




from collections import OrderedDict
import sys


def llama_2_7b(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = llama_2(architecture_dict=architecture_dict,
                               workload_dict=workload_dict)
    return performance_dict

def llama_2_13b(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = llama_2(architecture_dict=architecture_dict,
                               workload_dict=workload_dict)
    return performance_dict

def llama_2_70b(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = llama_2(architecture_dict=architecture_dict,
                               workload_dict=workload_dict)
    return performance_dict

def llama_2_70b_GQA(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = llama_2(architecture_dict=architecture_dict,
                               workload_dict=workload_dict)
    return performance_dict

def llama_2(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()

    gemm_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    nonlinear_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})

    performance_dict['subevent'] = OrderedDict({
        'gemm': gemm_dict,
        'nonlinear': nonlinear_dict
    })

    return performance_dict


import os, csv

from loguru import logger

from archx.utils import strip_list, read_yaml, create_dir


linear_interpolation_keywords = ['acc', 'add', 'sub', 'reg', 'rng', 'shifter']
quadratic_interpolation_keywords = ['multiplier']


tech_map = {
    'NanGate45': '45',
    'ASAP7': '7',
    'TNN7': '7',
    }


def extract(technology, frequency):
    """
    technology node name
    frequency in MHz
    """
    current_dir = os.path.dirname(os.path.abspath(__file__))
    rpt_dir = f'{current_dir}/syn_pnr_rpt/{technology}'
    csv_dir = f'{current_dir}/syn_pnr_csv/'
    create_dir(csv_dir)

    for module in os.listdir(rpt_dir):
        interpolation = 'linear'
        for keyword in linear_interpolation_keywords:
            if keyword in module:
                interpolation = 'linear'
        for keyword in quadratic_interpolation_keywords:
            if keyword in module:
                interpolation = 'quadratic'
        
        module_rpt = f'{rpt_dir}/{module}/{module}_DETAILS.rpt'

        # read from module_rpt
        with open(module_rpt, 'r') as file:
            for entry in file:
                elems = entry.strip().split(',')
                elems = strip_list(elems)
                if len(elems) > 0:
                    if str(elems[0]) == 'postRouteOpt':
                        area = float(elems[4]) / 10**6 # mm^2
                        leakage = float(elems[6]) # uW
                        dynamic = float(elems[7]) # uW

        assert dynamic != 0, logger.error(f'dynamic power={dynamic} for {module} is 0')
        assert leakage != 0, logger.error(f'leakage power={leakage} for {module} is 0')
        assert area != 0, logger.error(f'area={area} for {module} is 0')

        # parsing parameter dict
        header_ = []
        default_ = []
        param_dict = read_yaml(f'{current_dir}/param.yaml')
        if module in param_dict.keys() and param_dict[module] != None:
            for item_ in param_dict[module].items():
                header_.append(item_[0].lower())
                default_.append(item_[1])

        # generate csv table with parameters
        module_csv = csv_dir + module + '.csv'
        with open(module_csv, 'w') as f:
            csvwriter = csv.writer(f)
            header = ['technology', 'frequency', 'dynamic_uw', 'leakage_uw', 'area_mm2', 'num_instances', 'interpolation'] + header_
            csvwriter.writerow(header)
            content = [tech_map[technology], frequency, dynamic, leakage, area, 1, interpolation] + default_
            csvwriter.writerow(content)
    

if __name__ == '__main__':
    technology = 'TNN7'
    frequency = 100
    extract(technology, frequency)




from collections import OrderedDict
from zoo.llm.common.performance.mapping.mapping_performance import mapping
from zoo.llm.common.performance.utils import sum_subevents
from loguru import logger

# region functions
def gemm(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    layers = next(iter(workload_dict.values()))['configuration']['layers']

    projection_dict = OrderedDict({'count': layers, 'aggregation': 'sequential'})
    attention_dict = OrderedDict({'count': layers, 'aggregation': 'sequential'})
    ffn_dict = OrderedDict({'count': layers, 'aggregation': 'sequential'})
    output_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})

    performance_dict['subevent'] = OrderedDict({
        'projection': projection_dict,
        'attention': attention_dict,
        'ffn': ffn_dict,
        'output': output_dict
    })

    return performance_dict

def nonlinear(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    layers = next(iter(workload_dict.values()))['configuration']['layers']

    softmax_dict = OrderedDict({'count': layers, 'aggregation': 'sequential'})
    silu_dict = OrderedDict({'count': layers, 'aggregation': 'sequential'})

    performance_dict['subevent'] = OrderedDict({
        'softmax': softmax_dict,
        'silu': silu_dict
    })

    return performance_dict
# endregion

# region layers
def projection(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()

    proj_q_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    proj_k_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    proj_v_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    proj_a_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    
    performance_dict['subevent'] = OrderedDict({
        'proj_q': proj_q_dict,
        'proj_k': proj_k_dict,
        'proj_v': proj_v_dict,
        'proj_a': proj_a_dict
    })

    return performance_dict

def attention(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()

    qkt_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    av_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})

    performance_dict['subevent'] = OrderedDict({
        'qkt': qkt_dict,
        'av': av_dict
    })

    return performance_dict

def ffn(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()

    proj_up_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    proj_gate_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    proj_down_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})

    performance_dict['subevent'] = OrderedDict({
        'proj_up': proj_up_dict,
        'proj_gate': proj_gate_dict,
        'proj_down': proj_down_dict
    })

    return performance_dict

def output(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()

    output_prefill_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    output_decode_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})

    performance_dict['subevent'] = OrderedDict({
        'output_prefill': output_prefill_dict,
        'output_decode': output_decode_dict
    })

    return performance_dict
# endregion

# region sublayers
def proj_q(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()

    proj_q_prefill_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    proj_q_decode_dict = OrderedDict({'count': 4096, 'aggregation': 'sequential'})

    performance_dict['subevent'] = OrderedDict({
        'proj_q_prefill': proj_q_prefill_dict,
        'proj_q_decode': proj_q_decode_dict
    })

    return performance_dict

def proj_k(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()

    proj_k_prefill_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    proj_k_decode_dict = OrderedDict({'count': 4096, 'aggregation': 'sequential'})

    performance_dict['subevent'] = OrderedDict({
        'proj_k_prefill': proj_k_prefill_dict,
        'proj_k_decode': proj_k_decode_dict
    })

    return performance_dict

def proj_v(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()

    proj_v_prefill_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    proj_v_decode_dict = OrderedDict({'count': 4096, 'aggregation': 'sequential'})

    performance_dict['subevent'] = OrderedDict({
        'proj_v_prefill': proj_v_prefill_dict,
        'proj_v_decode': proj_v_decode_dict
    })

    return performance_dict

def proj_a(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()

    proj_a_prefill_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    proj_a_decode_dict = OrderedDict({'count': 4096, 'aggregation': 'sequential'})

    performance_dict['subevent'] = OrderedDict({
        'proj_a_prefill': proj_a_prefill_dict,
        'proj_a_decode': proj_a_decode_dict
    })

    return performance_dict

def qkt(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()

    qkt_prefill_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    qkt_decode_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})

    performance_dict['subevent'] = OrderedDict({
        'qkt_prefill': qkt_prefill_dict,
        'qkt_decode': qkt_decode_dict
    })

    return performance_dict

def av(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()

    av_prefill_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    av_decode_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})

    performance_dict['subevent'] = OrderedDict({
        'av_prefill': av_prefill_dict,
        'av_decode': av_decode_dict
    })

    return performance_dict

def proj_up(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()

    proj_up_prefill_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    proj_up_decode_dict = OrderedDict({'count': 4096, 'aggregation': 'sequential'})

    performance_dict['subevent'] = OrderedDict({
        'proj_up_prefill': proj_up_prefill_dict,
        'proj_up_decode': proj_up_decode_dict
    })

    return performance_dict

def proj_gate(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()

    proj_gate_prefill_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    proj_gate_decode_dict = OrderedDict({'count': 4096, 'aggregation': 'sequential'})

    performance_dict['subevent'] = OrderedDict({
        'proj_gate_prefill': proj_gate_prefill_dict,
        'proj_gate_decode': proj_gate_decode_dict
    })

    return performance_dict

def proj_down(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()

    proj_down_prefill_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    proj_down_decode_dict = OrderedDict({'count': 4096, 'aggregation': 'sequential'})

    performance_dict['subevent'] = OrderedDict({
        'proj_down_prefill': proj_down_prefill_dict,
        'proj_down_decode': proj_down_decode_dict
    })

    return performance_dict

def softmax(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()

    softmax_prefill_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    softmax_decode_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})

    performance_dict['subevent'] = OrderedDict({
        'softmax_prefill': softmax_prefill_dict,
        'softmax_decode': softmax_decode_dict
    })

    return performance_dict

def silu(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()

    silu_prefill_dict = OrderedDict({'count': 1, 'aggregation': 'sequential'})
    silu_decode_dict = OrderedDict({'count': 4096, 'aggregation': 'sequential'})

    performance_dict['subevent'] = OrderedDict({
        'silu_prefill': silu_prefill_dict,
        'silu_decode': silu_decode_dict
    })

    return performance_dict
# endregion

# region sublayer prefill/decode
def proj_q_prefill(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles_cycles': workload_config['proj_avg_early_termination_cycles']})

    # attn_shape = (batch_size, prefill_seq_len, dim)
    # weight_shape = (dim, dim)
    # output_shape = (batch_size, prefill_seq_len, dim)

    batch_size = workload_config['batch_size']
    prefill_seq_len = workload_config['prefill_seq_len']
    dim = workload_config['dim']
    
    mapping_dict = OrderedDict({
        'event': 'gemm',
        'batch': 1,
        'm': batch_size * prefill_seq_len,
        'k': dim,
        'n': dim,
    })

    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
    
    return performance_dict

def proj_q_decode(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['proj_avg_early_termination_cycles']})

    # attn_shape = (batch_size, 1, dim)
    # weight_shape = (dim, dim)
    # output_shape = (batch_size, 1, dim)

    batch_size = workload_config['batch_size']
    dim = workload_config['dim']

    mapping_dict = OrderedDict({
        'event': 'gemm',
        'batch': 1,
        'm': batch_size,
        'k': dim,
        'n': dim,
    })

    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
    
    return performance_dict

def proj_k_prefill(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['proj_avg_early_termination_cycles']})

    # attn_shape = (batch_size, prefill_seq_len, dim)
    # weight_shape = (dim, dim * kv_heads / heads)
    # output_shape = (batch_size, prefill_seq_len, dim * kv_heads / heads)

    batch_size = workload_config['batch_size']
    prefill_seq_len = workload_config['prefill_seq_len']
    dim = workload_config['dim']
    heads = workload_config['heads']
    kv_heads = workload_config['kv_heads']

    mapping_dict = OrderedDict({
        'event': 'gemm',
        'batch': 1,
        'm': batch_size * prefill_seq_len,
        'k': dim,
        'n': dim * kv_heads / heads,
    })

    performance_dict = performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
    return performance_dict

def proj_k_decode(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['proj_avg_early_termination_cycles']})

    # attn_shape = (batch_size, 1, dim)
    # weight_shape = (dim, dim * kv_heads / heads)
    # output_shape = (batch_size, 1, dim * kv_heads / heads)

    batch_size = workload_config['batch_size']
    dim = workload_config['dim']
    heads = workload_config['heads']
    kv_heads = workload_config['kv_heads']

    mapping_dict = OrderedDict({
        'event': 'gemm',
        'batch': 1,
        'm': batch_size,
        'k': dim,
        'n': dim * kv_heads / heads,
    })

    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
    return performance_dict

def proj_v_prefill(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['proj_avg_early_termination_cycles']})

    # attn_shape = (batch_size, prefill_seq_len, dim)
    # weight_shape = (dim, dim * kv_heads / heads)
    # output_shape = (batch_size, prefill_seq_len, dim * kv_heads / heads)

    batch_size = workload_config['batch_size']
    prefill_seq_len = workload_config['prefill_seq_len']
    dim = workload_config['dim']
    heads = workload_config['heads']
    kv_heads = workload_config['kv_heads']

    mapping_dict = OrderedDict({
        'event': 'gemm',
        'batch': 1,
        'm': batch_size * prefill_seq_len,
        'k': dim,
        'n': dim * kv_heads / heads,
    })

    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
    return performance_dict

def proj_v_decode(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['proj_avg_early_termination_cycles']})

    # attn_shape = (batch_size, 1, dim)
    # weight_shape = (dim, dim * kv_heads / heads)
    # output_shape = (batch_size, 1, dim * kv_heads / heads)

    batch_size = workload_config['batch_size']
    dim = workload_config['dim']
    heads = workload_config['heads']
    kv_heads = workload_config['kv_heads']

    mapping_dict = OrderedDict({
        'event': 'gemm',
        'batch': 1,
        'm': batch_size,
        'k': dim,
        'n': dim * kv_heads / heads,
    })
    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
    return performance_dict

def proj_a_prefill(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['proj_avg_early_termination_cycles']})

    # attn_shape = (batch_size, prefill_seq_len, dim)
    # weight_shape = (dim, dim)
    # output_shape = (batch_size, prefill_seq_len, dim)

    batch_size = workload_config['batch_size']
    prefill_seq_len = workload_config['prefill_seq_len']
    dim = workload_config['dim']

    mapping_dict = OrderedDict({
        'event': 'gemm',
        'batch': 1,
        'm': batch_size * prefill_seq_len,
        'k': dim,
        'n': dim,
    })

    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
    return performance_dict

def proj_a_decode(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['proj_avg_early_termination_cycles']})

    # attn_shape = (batch_size, 1, dim)
    # weight_shape = (dim, dim)
    # output_shape = (batch_size, 1, dim)

    batch_size = workload_config['batch_size']
    dim = workload_config['dim']

    mapping_dict = OrderedDict({
        'event': 'gemm',
        'batch': 1,
        'm': batch_size,
        'k': dim,
        'n': dim,
    })

    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
    return performance_dict

def qkt_prefill(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['k_avg_early_termination_cycles']})

    # attn_shape = (batch_size, heads, prefill_seq_len, dim/heads)
    # weight_shape = (batch_size, heads, dim/heads, prefill_seq_len)
    # output_shape = (batch_size, heads, prefill_seq_len, prefill_seq_len)

    # with gqa
    # attn_shape = (batch_size, kv_heads, heads / kv_heads * prefill_seq_len, dim/heads)
    # weight_shape = (batch_size, kv_heads, dim/heads, prefill_seq_len)
    # output_shape = (batch_size, kv_heads, heads / kv_heads * prefill_seq_len, prefill_seq_len)

    batch_size = workload_config['batch_size']
    heads = workload_config['heads']
    prefill_seq_len = workload_config['prefill_seq_len']
    dim = workload_config['dim']
    kv_heads = workload_config['kv_heads']

    mapping_dict = OrderedDict({
        'event': 'gemm',
        'batch': batch_size * kv_heads,
        'm': (heads / kv_heads) * prefill_seq_len,
        'k': dim/heads,
        'n': prefill_seq_len,
    })

    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
    return performance_dict

def qkt_decode(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['k_avg_early_termination_cycles']})

    # attn_shape = (batch_size, heads, 1, dim/heads)
    # weight_shape = (batch_size, heads, dim/heads, cur_seq_len)
    # output_shape = (batch_size, heads, 1, cur_seq_len)

    # with gqa
    # attn_shape = (batch_size, kv_heads, heads / kv_heads, dim/heads)
    # weight_shape = (batch_size, kv_heads, dim/heads, cur_seq_len)
    # output_shape = (batch_size, kv_heads, heads / kv_heads, cur_seq_len)

    batch_size = workload_config['batch_size']
    heads = workload_config['heads']
    prefill_seq_len = workload_config['prefill_seq_len']
    max_seq_len = workload_config['max_seq_len']
    dim = workload_config['dim']
    kv_heads = workload_config['kv_heads']

    for cur_seq_len in range(prefill_seq_len + 1, max_seq_len + 1):
        mapping_dict = OrderedDict({
            'event': 'gemm',
            'batch': batch_size * kv_heads,
            'm': (heads / kv_heads),
            'k': dim/heads,
            'n': cur_seq_len,
        })
        if not performance_dict:
            performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
        else:
            performance_dict = sum_subevents(performance_dict, mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config))

    return performance_dict

def av_prefill(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['v_avg_early_termination_cycles']})

    # attn_shape = (batch_size, heads, 1, prefill_seq_len)
    # weight_shape = (batch_size, heads, prefill_seq_len, dim/heads)
    # output_shape = (batch_size, heads, 1, dim/heads)

    # with gqa
    # attn_shape = (batch_size, kv_heads, heads / kv_heads * prefill_seq_len, prefill_seq_len)
    # weight_shape = (batch_size, kv_heads, prefill_seq_len, dim/heads)
    # output_shape = (batch_size, kv_heads, heads / kv_heads * prefill_seq_len, dim/heads)

    batch_size = workload_config['batch_size']
    heads = workload_config['heads']
    kv_heads = workload_config['kv_heads']
    prefill_seq_len = workload_config['prefill_seq_len']
    dim = workload_config['dim']

    mapping_dict = OrderedDict({
        'event': 'gemm',
        'batch': batch_size * kv_heads,
        'm': (heads / kv_heads) * prefill_seq_len,
        'k': prefill_seq_len,
        'n': dim/heads,
    })


    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)

    return performance_dict

def av_decode(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']
    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['v_avg_early_termination_cycles']})

    # attn_shape = (batch_size, heads, 1, cur_seq_len)
    # weight_shape = (batch_size, heads, cur_seq_len, dim/heads)
    # output_shape = (batch_size, heads, 1, dim/heads)

    # with gqa
    # attn_shape = (batch_size, kv_heads, heads / kv_heads, cur_seq_len)
    # weight_shape = (batch_size, kv_heads, cur_seq_len, dim/heads)
    # output_shape = (batch_size, kv_heads, heads / kv_heads, dim/heads)

    batch_size = workload_config['batch_size']
    heads = workload_config['heads']
    kv_heads = workload_config['kv_heads']
    max_seq_len = workload_config['max_seq_len']
    prefill_seq_len = workload_config['prefill_seq_len']
    dim = workload_config['dim']

    for cur_seq_len in range(prefill_seq_len + 1, max_seq_len + 1):
        mapping_dict = OrderedDict({
            'event': 'gemm',
            'batch': batch_size * kv_heads,
            'm': (heads / kv_heads),
            'k': cur_seq_len,
            'n': dim/heads,
        })

        if not performance_dict:
            performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
        else:
            performance_dict = sum_subevents(performance_dict, mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config))
  

    return performance_dict

def proj_up_prefill(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['ffn_avg_early_termination_cycles']})

    # attn_shape = (batch_size, prefill_seq_len, dim)
    # weight_shape = (dim, hidden_dim)
    # output_shape = (batch_size, prefill_seq_len, hidden_dim)

    batch_size = workload_config['batch_size']
    prefill_seq_len = workload_config['prefill_seq_len']
    dim = workload_config['dim']
    hidden_dim = workload_config['hidden_dim']

    mapping_dict = OrderedDict({
        'event': 'gemm',
        'batch': 1,
        'm': batch_size * prefill_seq_len,
        'k': dim,
        'n': hidden_dim,
    })

    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
    return performance_dict

def proj_up_decode(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['ffn_avg_early_termination_cycles']})

    # attn_shape = (batch_size, 1, dim)
    # weight_shape = (dim, hidden_dim)
    # output_shape = (batch_size, 1, hidden_dim)

    batch_size = workload_config['batch_size']
    dim = workload_config['dim']
    hidden_dim = workload_config['hidden_dim']

    mapping_dict = OrderedDict({
        'event': 'gemm',
        'batch': 1,
        'm': batch_size,
        'k': dim,
        'n': hidden_dim,
    })

    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
    return performance_dict

def proj_gate_prefill(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['ffn_avg_early_termination_cycles']})

    # attn_shape = (batch_size, prefill_seq_len, dim)
    # weight_shape = (dim, hidden_dim)
    # output_shape = (batch_size, prefill_seq_len, hidden_dim)

    batch_size = workload_config['batch_size']
    prefill_seq_len = workload_config['prefill_seq_len']
    dim = workload_config['dim']
    hidden_dim = workload_config['hidden_dim']

    mapping_dict = OrderedDict({
        'event': 'gemm',
        'batch': 1,
        'm': batch_size * prefill_seq_len,
        'k': dim,
        'n': hidden_dim,
    })

    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
    return performance_dict

def proj_gate_decode(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['ffn_avg_early_termination_cycles']})

    # attn_shape = (batch_size, 1, dim)
    # weight_shape = (dim, hidden_dim)
    # output_shape = (batch_size, 1, hidden_dim)

    batch_size = workload_config['batch_size']
    dim = workload_config['dim']
    hidden_dim = workload_config['hidden_dim']

    mapping_dict = OrderedDict({
        'event': 'gemm',
        'batch': 1,
        'm': batch_size,
        'k': dim,
        'n': hidden_dim,
    })

    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
    return performance_dict

def proj_down_prefill(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['ffn_avg_early_termination_cycles']})

    # attn_shape = (batch_size, prefill_seq_len, hidden_dim)
    # weight_shape = (hidden_dim, dim)
    # output_shape = (batch_size, prefill_seq_len, dim)

    batch_size = workload_config['batch_size']
    prefill_seq_len = workload_config['prefill_seq_len']
    hidden_dim = workload_config['hidden_dim']
    dim = workload_config['dim']

    mapping_dict = OrderedDict({
        'event': 'gemm',
        'batch': 1,
        'm': batch_size * prefill_seq_len,
        'k': hidden_dim,
        'n': dim,
    })

    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
    return performance_dict

def proj_down_decode(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['ffn_avg_early_termination_cycles']})

    # attn_shape = (batch_size, 1, hidden_dim)
    # weight_shape = (hidden_dim, dim)
    # output_shape = (batch_size, 1, dim)

    batch_size = workload_config['batch_size']
    hidden_dim = workload_config['hidden_dim']
    dim = workload_config['dim']

    mapping_dict = OrderedDict({
        'event': 'gemm',
        'batch': 1,
        'm': batch_size,
        'k': hidden_dim,
        'n': dim,
    })

    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
    return performance_dict

def softmax_prefill(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['nonlinear_avg_early_termination_cycles']})

    # attn_shape = (batch_size, heads, prefill_seq_len, prefill_seq_len)
    # output_shape = (batch_size, heads, prefill_seq_len, prefill_seq_len)

    # with gqa
    # attn_shape = (batch_size, kv_heads, heads / kv_heads * prefill_seq_len, cur_seq_len)
    # output_shape = (batch_size, kv_heads, heads / kv_heads * prefill_seq_len, cur_seq_len)

    batch_size = workload_config['batch_size']
    heads = workload_config['heads']
    prefill_seq_len = workload_config['prefill_seq_len']

    mapping_dict = OrderedDict({
        'event': 'nonlinear',
        'function': 'softmax',
        'batch': 1,
        'm': batch_size * heads * prefill_seq_len,
        'n': prefill_seq_len,
    })

    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
    return performance_dict

def softmax_decode(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['nonlinear_avg_early_termination_cycles']})

    # attn_shape = (batch_size, heads, 1, cur_seq_len)
    # output_shape = (batch_size, heads, 1, cur_seq_len)

    # with gqa
    # attn_shape = (batch_size, kv_heads, heads / kv_heads, cur_seq_len)
    # output_shape = (batch_size, kv_heads, heads / kv_heads, cur_seq_len)

    batch_size = workload_config['batch_size']
    kv_heads = workload_config['kv_heads']
    heads = workload_config['heads']
    max_seq_len = workload_config['max_seq_len']
    prefill_seq_len = workload_config['prefill_seq_len']

    test = batch_size * heads * prefill_seq_len
    for cur_seq_len in range(prefill_seq_len + 1, max_seq_len + 1):
        test += batch_size * heads * cur_seq_len
        mapping_dict = OrderedDict({
            'event': 'nonlinear',
            'function': 'softmax',
            'batch': 1,
            'm': batch_size * heads,
            'n': cur_seq_len,
        })

        if not performance_dict:
            performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
        else:
            performance_dict = sum_subevents(performance_dict, mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config))
    
    return performance_dict

def silu_prefill(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['nonlinear_avg_early_termination_cycles']})

    # attn_shape = (batch_size, prefill_seq_len, hidden_dim)
    # output_shape = (batch_size, prefill_seq_len, hidden_dim)

    batch_size = workload_config['batch_size']
    prefill_seq_len = workload_config['prefill_seq_len']
    hidden_dim = workload_config['hidden_dim']

    mapping_dict = OrderedDict({
        'event': 'nonlinear',
        'function': 'silu',
        'batch': 1,
        'm': hidden_dim,
        'n': batch_size * prefill_seq_len,
    })

    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
    return performance_dict

def silu_decode(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['nonlinear_avg_early_termination_cycles']})

    # attn_shape = (batch_size, 1, hidden_dim)
    # output_shape = (batch_size, 1, hidden_dim)

    batch_size = workload_config['batch_size']
    hidden_dim = workload_config['hidden_dim']

    mapping_dict = OrderedDict({
        'event': 'nonlinear',
        'function': 'silu',
        'batch': 1,
        'm': hidden_dim,
        'n': batch_size,
    })

    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)

    return performance_dict

def output_prefill(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['default_avg_early_termination_cycles']})

    # attn_shape = (batch_size, prefill_seq_len, dim)
    # weight_shape = (dim, vocab_size)
    # output_shape = (batch_size, prefill_seq_len, vocab_size)

    batch_size = workload_config['batch_size']
    prefill_seq_len = workload_config['prefill_seq_len']
    dim = workload_config['dim']
    vocab_size = workload_config['vocab_size']

    mapping_dict = OrderedDict({
        'event': 'gemm',
        'batch': 1,
        'm': batch_size * prefill_seq_len,
        'k': dim,
        'n': vocab_size,
    })

    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
    return performance_dict

def output_decode(architecture_dict: OrderedDict, workload_dict: OrderedDict=None) -> OrderedDict:
    performance_dict = OrderedDict()
    workload_config = next(iter(workload_dict.values()))['configuration']

    if workload_config['architecture'].lower() == 'mugi':
        workload_config.update({'early_termination_cycles': workload_config['default_avg_early_termination_cycles']})

    # attn_shape = (batch_size, 1, dim)
    # weight_shape = (dim, vocab_size)
    # output_shape = (batch_size, 1, vocab_size)

    batch_size = workload_config['batch_size']
    dim = workload_config['dim']
    vocab_size = workload_config['vocab_size']

    mapping_dict = OrderedDict({
        'event': 'gemm',
        'batch': 1,
        'm': batch_size,
        'k': dim,
        'n': vocab_size,
    })

    performance_dict = mapping(mapping_dict=mapping_dict, architecture_dict=architecture_dict, workload_dict=workload_config)
    return performance_dict
# endregion


import importlib, sys
import graph_tool.all as gt

from collections import OrderedDict
from loguru import logger

from archx.utils import get_path


key_aggregation = 'aggregation'
key_operation = 'operation'
legal_specified = ['parallel', 'sequential']
key_value = 'value'
key_unit = 'unit'
key_subevent = 'subevent'
key_count = 'count'
key_factor = 'factor'


def import_function_from_path(file_path: str, function: str) -> callable:
    full_path = get_path(file_path)
    
    spec = importlib.util.spec_from_file_location(function, full_path)
    module_py = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module_py
    spec.loader.exec_module(module_py)

    # get the specified function dynamically
    if hasattr(module_py, function) and callable(getattr(module_py, function)):
        target_function = getattr(module_py, function)
        return target_function
    else:
        logger.error(f'Invalid function <{function}> at <{full_path}>.')
        exit()
    

def simulate_performance_one_event(event_graph: gt.Graph, architecture_dict: OrderedDict, workload_dict: OrderedDict, event_name: str) -> gt.Graph:
    # run performance model for a single event node
    v = gt.find_vertex(event_graph, event_graph.vp.event, event_name)
    assert len(v) > 0, logger.error(f'Invalid event <{event_name}>.')

    v = v[0]
    performance_path = event_graph.vp.performance[v]
    if performance_path is None or performance_path == 'None':
        assert v.out_degree() == 0, logger.error(f'Missing performance model for event <{event_name}>.')
        logger.info(f'Module <{event_name}> has no performance model.')
    else:
        # if the current node is not a leaf node, update edges with performance model
        performance_model = import_function_from_path(performance_path, function=event_name)
        performance_dict = performance_model(architecture_dict=architecture_dict, workload_dict=workload_dict)
        assert performance_dict is not None, logger.error(f'No performance model returned for event <{event_name}>')
        # process additional metrics in specified mode
        if len(list(performance_dict.keys())) > 1:
            metric_key_list = list(performance_dict.keys())
            metric_key_list.remove(key_subevent)
            for metric_key in metric_key_list:
                if metric_key not in event_graph.vp:
                    event_graph.vp[metric_key] = event_graph.new_vertex_property('object')
                event_graph.vp[metric_key][v] = OrderedDict({key_value: performance_dict[metric_key][key_value], key_unit: performance_dict[metric_key][key_unit]})

        # iterate over all out-edges
        for e in v.out_edges():
            edge_source = event_graph.vp.event[e.source()]
            edge_target = event_graph.vp.event[e.target()]
            assert edge_target in performance_dict[key_subevent], logger.error(f'  Missing subevent <{edge_target}> in the <{edge_source}> performance model.')
            if key_count not in performance_dict[key_subevent][edge_target]:
                event_graph.ep.count[e] = 1.
                logger.warning(f'  Missing count in subevent <{edge_target}>; default to 1.')
            else:
                event_graph.ep.count[e] = performance_dict[key_subevent][edge_target][key_count]

            # default specified aggregation is parallel, which will be reduced with maximum
            if key_aggregation in performance_dict[key_subevent][edge_target]:
                assert performance_dict[key_subevent][edge_target][key_aggregation].lower() in legal_specified, logger.error(f'  Invalid aggregation <{performance_dict["subevent"][edge_target][key_aggregation]}> in the performance model of event <{event_name}> at <{performance_path}>; legal values: {legal_specified}.')
                event_graph.ep.aggregation[e] = performance_dict[key_subevent][edge_target][key_aggregation].lower()
                if e.target().out_degree() == 0:
                    logger.warning(f'  Ignore aggregation <{event_graph.ep.aggregation[e]}> between event <{edge_source}> and module <{edge_target}>; aggregation only takes effect between events.')
            else:
                logger.warning(f'  Missing aggregation in subevent <{edge_target}>; default to <parallel>.')
                event_graph.ep.aggregation[e] = 'parallel'

            # default operation is empty, in which case, query is a direct dictionary lookup
            if key_operation in performance_dict[key_subevent][edge_target]:
                assert e.target().out_degree() == 0, logger.error(f'  Invalid operation between event <{event_name}> and event <{event_graph.vp.event[e.target()]}>; operation should be between event and module.')
                event_graph.ep.operation[e] = performance_dict[key_subevent][edge_target][key_operation]
            
            # default factor is an empty dict, which is used to scale queried results
            if key_factor in performance_dict[key_subevent][edge_target]:
                event_graph.ep.factor[e] = performance_dict[key_subevent][edge_target][key_factor]
                assert isinstance(event_graph.ep.factor[e], dict), logger.error(f'  Invalid factor <{event_graph.ep.factor[e]}> between event <{event_name}> and module <{event_graph.vp.event[e.target()]}>; factor should be a dict.')

            logger.debug(f'  Event <{edge_source}> has <{event_graph.ep.count[e]}> subevent <{edge_target}> with specified aggregation <{event_graph.ep.aggregation[e]}>.')

        logger.success(f'Simulate event <{event_name}> at <{performance_path}>.')

    return event_graph


def simulate_performance_all_events(event_graph: gt.Graph, architecture_dict: OrderedDict, workload_dict: OrderedDict) -> gt.Graph:
    # Iterate over all event nodes
    for v in event_graph.vertices():
        event_name = event_graph.vp.event[v]
        event_graph = simulate_performance_one_event(event_graph, architecture_dict, workload_dict, event_name)
    
    logger.success(f'Simulate all events.')
    
    return event_graph




import subprocess, os, time, random, copy

from collections import OrderedDict

from loguru import logger


def cacti7_run(
    mem_type: str,
    tech_node_nm: str,
    cfg_dict: OrderedDict,
    origin_cfg_file=None,
    target_cfg_file=None,
    result_file=None
):
    """
    run the cacti with input configuration, work for SRAM, whose size is either calculated to match the bw or pre-specified
    """
    original = open(origin_cfg_file, 'r')
    target   = open(target_cfg_file, 'w')

    tech_node_um = float(int(tech_node_nm)/1000)

    if mem_type == 'sram':
        bank_count = int(cfg_dict['bank'])
        block_bits = int(cfg_dict['width'])
        block_bytes = int(cfg_dict['width'] / 8)
        total_bytes = int(bank_count * cfg_dict['depth'] * block_bytes)

        target.write('-size (bytes) ' + str(total_bytes) + '\n')
        target.write('-block size (bytes) ' + str(block_bytes) + '\n')
        target.write('-technology (u) ' + str(tech_node_um) + '\n')
        target.write('-UCA bank count '+ str(bank_count) + '\n')
        target.write('-output/input bus width ' + str(block_bits) + '\n')

    elif mem_type in ['dram', 'ddr4']:
        total_bytes = int(cfg_dict['size'] / 8)
        page_bits = 8192
        bank_count = 8
        prefetch = 8

        target.write('-size (bytes) ' + str(total_bytes) + '\n')
        target.write('-page size (bits) ' + str(page_bits) + '\n')
        target.write('-technology (u) ' + str(tech_node_um) + '\n')
        target.write('-UCA bank count '+ str(bank_count) + '\n')
        target.write('-internal prefetch width '+ str(prefetch) + '\n')
    
    for entry in original:
        target.write(entry)
    
    original.close()
    target.close()

    run_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'include/cacti7')
    if not os.path.exists(os.path.join(run_dir, 'cacti')):
        subprocess.call(['make', 'all'], shell=True, cwd=run_dir)
        time.sleep(10)
    
    tartget_cacti_name = result_file.replace('//', '/').replace('/', '_').split('.')[0] + '_'
    tartget_cacti_name += str(time.time())

    rep_cmd = 'cp ./cacti ./cacti_' + tartget_cacti_name
    subprocess.call([rep_cmd], shell=True, cwd=run_dir)
    final_cmd = './cacti_' + tartget_cacti_name + ' -infile ' + target_cfg_file + ' > ' + result_file
    subprocess.call([final_cmd], shell=True, cwd=run_dir)
    rm_cmd = 'rm -rf ./cacti_' + tartget_cacti_name
    subprocess.call([rm_cmd], shell=True, cwd=run_dir)


def parse_report_line_count(
    report=None,
    index=64
):
    # get the area and power numbers in the report for final memory power and energy estimation.
    cacti_out = open(report, 'r')
    line_idx = 0
    for entry in cacti_out:
        line_idx += 1
    broken_rpt = line_idx <= index
    return broken_rpt


def parse_report_sram(
    report=None
):
    # get the area and power numbers in the report for final memory power and energy estimation.
    cacti_out = open(report, 'r')

    line_idx = 0
    for entry in cacti_out:
        line_idx += 1
        if line_idx == 12:
            ram_type = entry.strip().split(':')[-1].strip()
            assert ram_type == 'Scratch RAM', 'Invalid SRAM type.'
        if line_idx == 50:
            # unit: ns
            bank = float(entry.strip().split(':')[-1].strip())
        if line_idx == 58:
            # unit: ns
            access_time = float(entry.strip().split(':')[-1].strip())
        if line_idx == 59:
            # unit: ns
            cycle_time = float(entry.strip().split(':')[-1].strip())
        if line_idx == 60:
            # unit: nJ
            dynamic_energy_rd = float(entry.strip().split(':')[-1].strip())
        if line_idx == 61:
            # unit: nJ
            dynamic_energy_wr = float(entry.strip().split(':')[-1].strip())
        if line_idx == 62:
            # unit: mW
            leakage_power_bank = float(entry.strip().split(':')[-1].strip())
        if line_idx == 63:
            # unit: mW
            gate_leakage_power_bank = float(entry.strip().split(':')[-1].strip())
        if line_idx == 64:
            # unit: mm^2
            height = float(entry.strip().split(':')[-1].split('x')[0].strip())
            width = float(entry.strip().split(':')[-1].split('x')[1].strip())
            area = height * width

        if line_idx == 2:
            # unit: byte
            block_sz_bytes = float(entry.strip().split(':')[-1].strip())
    
    broken_rpt = parse_report_line_count(report, 64)
    assert not broken_rpt, logger.error('Check ' + report + ' for sram cacti failure.')

    # MHz
    max_freq = 1 / cycle_time * 1000
    # nJ
    energy_per_rd_nJ = dynamic_energy_rd
    # nJ
    energy_per_wr_nJ = dynamic_energy_wr
    # mW
    leakage_power_mW = (leakage_power_bank + gate_leakage_power_bank) * bank
    # mm^2
    total_area_mm2 = area
    cacti_out.close()
    return energy_per_rd_nJ, energy_per_wr_nJ, leakage_power_mW, total_area_mm2, block_sz_bytes, max_freq


def parse_report_dram(
    report=None
):
    # get the area and power numbers in the report for final memory power and energy estimation.
    cacti_out = open(report, 'r')

    line_idx = 0
    for entry in cacti_out:
        line_idx += 1
        if line_idx == 12:
            ram_type = entry.strip().split(':')[-1].strip()
            assert ram_type == 'Scratch RAM', 'Invalid DRAM type.'
        if line_idx == 50:
            # unit: ns
            bank = float(entry.strip().split(':')[-1].strip())
        if line_idx == 58:
            # unit: ns
            access_time = float(entry.strip().split(':')[-1].strip())
        if line_idx == 59:
            # unit: ns
            cycle_time = float(entry.strip().split(':')[-1].strip())
        if line_idx == 61:
            # unit: nJ
            activate_energy = float(entry.strip().split(':')[-1].strip())
        if line_idx == 62:
            # unit: nJ
            energy_rd = float(entry.strip().split(':')[-1].strip())
        if line_idx == 63:
            # unit: nJ
            energy_wr = float(entry.strip().split(':')[-1].strip())
        if line_idx == 64:
            # unit: nJ
            precharge_energy = float(entry.strip().split(':')[-1].strip())
        if line_idx == 65:
            # unit: mW
            leakage_power_closed_page = float(entry.strip().split(':')[-1].strip())
        if line_idx == 66:
            # unit: mW
            leakage_power_open_page = float(entry.strip().split(':')[-1].strip())
        if line_idx == 67:
            # unit: mW
            leakage_power_IO = float(entry.strip().split(':')[-1].strip())
        if line_idx == 68:
            # unit: mW
            refresh_power = float(entry.strip().split(':')[-1].strip())
        if line_idx == 69:
            # unit: mm^2
            height = float(entry.strip().split(':')[-1].split('x')[0].strip())
            width = float(entry.strip().split(':')[-1].split('x')[1].strip())
            area = height * width

    broken_rpt = parse_report_line_count(report, 69)
    assert not broken_rpt, logger.error('Check ' + report + ' for dram cacti failure.')

    # MHz
    max_freq = 1 / cycle_time * 1000
    cacti_out.close()
    return activate_energy, energy_rd, energy_wr, precharge_energy, leakage_power_closed_page, leakage_power_open_page, leakage_power_IO, refresh_power, area, max_freq


def query(name: str, interface: str, query: OrderedDict, input_dir=None, output_dir=None):
    query = copy.deepcopy(query)
    query_class = query['class'].lower()
    query_technology = query['technology']
    query_frequency = query['frequency']
    del query['class']
    del query['technology']
    del query['frequency']

    assert os.path.isdir(output_dir), logger.error(f'Invalid output dir <{output_dir}>.')

    origin_cfg_file = os.path.dirname(os.path.abspath(__file__))
    assert query_class in ['sram', 'dram', 'ddr4'], logger.error(f'Invalid <{query_class}> in CACTI7; valid values: [sram, dram, ddr4]')
    if query_class == 'sram':
        origin_cfg_file = origin_cfg_file + '/sram.cfg'
    elif query_class in ['dram', 'ddr4']:
        origin_cfg_file = origin_cfg_file + '/ddr4.cfg'
    assert os.path.isfile(origin_cfg_file), logger.error(f'Invalid CACTI7 config file <{origin_cfg_file}>.')
    
    if query_class == 'sram':
        post_fix = '.sram' + '.width' + str(query['width']) + '.depth' + str(query['depth']) + '.bank' + str(query['bank'])
    elif query_class in ['dram', 'ddr4']:
        post_fix = '.ddr4' + '.size' + str(query['size'])

    target_cfg_file = os.path.join(output_dir, name + post_fix + '.cacti7.cfg')
    cacti_report = os.path.join(output_dir, name + post_fix + '.cacti7.rpt')

    target_cfg_file_flag = target_cfg_file + '.flag'
    if not os.path.exists(target_cfg_file_flag):
        cacti7_run(query_class, query_technology, query, origin_cfg_file, target_cfg_file, cacti_report)
        f = open(target_cfg_file_flag, 'a')
        f.write(target_cfg_file_flag)
        f.close()
    else:
        broken_rpt = parse_report_line_count(cacti_report, 64)
        if broken_rpt:
            os.remove(cacti_report)
            os.remove(target_cfg_file_flag)
            cacti7_run(query_class, query_technology, query, origin_cfg_file, target_cfg_file, cacti_report)

    if query_class == 'sram':
        energy_per_rd_nJ, energy_per_wr_nJ, leakage_power_mW, total_area_mm2, _, _ = parse_report_sram(cacti_report)
    elif query_class in ['dram', 'ddr4']:
        activate_energy, energy_rd, energy_wr, precharge_energy, leakage_power_closed_page, leakage_power_open_page, leakage_power_IO, refresh_power, area, max_freq = parse_report_dram(cacti_report)
        
        if 'embedded' not in query or query['embedded'] is False:
            IO_power_factor = 1.
        else:
            IO_power_factor = 0.

        bank_count = 8
        constant_power = (leakage_power_IO * IO_power_factor + refresh_power) * bank_count

        if 'open_page' not in query or query['open_page'] is False:
            constant_power = constant_power + leakage_power_closed_page * bank_count
        else:
            # default to open page
            constant_power = constant_power + leakage_power_open_page * bank_count
        
        energy_per_rd_nJ = energy_rd
        energy_per_wr_nJ = energy_wr
        leakage_power_mW = constant_power
        total_area_mm2 = area

    output_dict = OrderedDict()
    output_dict['technology'] = OrderedDict({'value': query_technology, 'unit': 'nm'})
    output_dict['frequency'] = OrderedDict({'value': query_frequency, 'unit': 'MHz'})
    output_dict['dynamic_energy'] = OrderedDict({'read': OrderedDict({'value': energy_per_rd_nJ, 'unit': 'nJ'}),
                                                 'write': OrderedDict({'value': energy_per_wr_nJ, 'unit': 'nJ'})})
    output_dict['leakage_power'] = OrderedDict({'value': leakage_power_mW, 'unit': 'mW'})
    output_dict['area'] = OrderedDict({'value': total_area_mm2, 'unit': 'mm^2'})

    return output_dict


if __name__ == '__main__':
    pass



# following two lines are used in testing
import sys, os, shutil

import graph_tool.all as gt

from loguru import logger

from archx.event import create_event_graph, save_event_graph, load_event_graph
from archx.utils import get_path, create_dir


def test_create_event_graph():
    event_file = 'examples/mac_1_cycle/input/event/example.event.yaml'
    event_graph = create_event_graph(event_file)

    create_dir('tests/test_event/')
    
    gt.graph_draw(event_graph, 
            vertex_text = event_graph.vp.event, 
            vertex_font_size=10, 
            edge_pen_width = event_graph.ep.count,
            output_size=(800, 800),
            output='tests/test_event/weighted_graph.pdf')

    ckpt_file = 'tests/test_event/example.event.gt'
    save_event_graph(event_graph, ckpt_file)
    event_graph_ckpt = load_event_graph(ckpt_file)

    gt.graph_draw(event_graph, 
            vertex_text = event_graph.vp.event, 
            vertex_font_size=10, 
            edge_pen_width = event_graph.ep.count,
            output_size=(800, 800),
            output='tests/test_event/weighted_graph_ckpt.pdf')

    # check whether two graphs are isomorphic
    assert gt.isomorphism(event_graph, event_graph_ckpt)


def test_cleanup():
    path = get_path('tests/test_event/')
    shutil.rmtree(path)


if __name__ == "__main__":
    test_create_event_graph()
    test_cleanup()




from zoo.llm.results.query.utils import query_execution_time, compute_latency, load_yaml
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import pandas as pd
import numpy as np
import os

def query(input_path, output_path):
    vlp_list = ['mugi', 'carat']
    vlp_arch_dim_list = ['128x8', '256x8']
    gemm_vlp_latency_module = 'and_gate'

    mugi_subarch_list = ['vlp', 'lut']

    baseline_list = ['systolic', 'simd']
    baseline_arch_dim_list = ['16x16']
    simd_subarch_list = ['mac', 'figna']
    systolic_subarch_list = ['mac', 'figna', 'taylor', 'pwl']
    gemm_baseline_latency_module = 'multiplier'

    mugi_nonlinear_module = 'magnitude_register'
    baseline_nonlinear_module = 'accumulator_vector'

    model_list = ['llama_2_7b', 'llama_2_13b', 'llama_2_70b', 'llama_2_70b_GQA']
    max_seq_len = 'max_seq_len_4096'
    batch_size = 'batch_size_8'
    network = 'single_node'

    end_to_end_breakdown = pd.DataFrame()

    for arch in vlp_list + baseline_list:
       for arch_dim in (vlp_arch_dim_list if arch in vlp_list else baseline_arch_dim_list if arch in baseline_list else ['8x16x16'] if arch in ['tensor'] else ['']):
            for subarch in (systolic_subarch_list if arch in ['systolic'] else simd_subarch_list if arch in ['simd'] else mugi_subarch_list if arch in ['mugi'] else ['']):
                for model in model_list:

                    module = gemm_vlp_latency_module if arch in vlp_list else gemm_baseline_latency_module
                    nonlinear_module = mugi_nonlinear_module if arch in ['mugi'] else baseline_nonlinear_module
                    termination_path = 'full_termination' if arch == 'mugi' else ''
                    run_path = os.path.normpath(f'{input_path}{arch}/{network}/{subarch}/{arch_dim}/{model}/{max_seq_len}/{batch_size}/{termination_path}/')
                    yaml_dict = load_yaml(run_path)

                    event_graph = yaml_dict['event_graph']
                    metric_dict = yaml_dict['metric_dict']

                    proj_execution_time = query_execution_time(event_graph=event_graph, metric_dict=metric_dict, workload=model, event='projection')
                    proj_latency_eff_dict = {
                        'latency': proj_execution_time
                    }

                    attn_execution_time = query_execution_time(event_graph=event_graph, metric_dict=metric_dict, workload=model, event='attention')
                    attn_latency_eff_dict = {
                        'latency': attn_execution_time
                    }

                    ffn_execution_time = query_execution_time(event_graph=event_graph, metric_dict=metric_dict, workload=model, event='ffn')
                    ffn_latency_eff_dict = {
                        'latency': ffn_execution_time
                    }

                    nonlinear_execution_time = query_execution_time(event_graph=event_graph, metric_dict=metric_dict, workload=model, event='nonlinear')
                    nonlinear_latency_eff_dict = {
                        'latency': nonlinear_execution_time
                    }

                    proj_metric_df = pd.DataFrame(proj_latency_eff_dict, index=[0])
                    proj_metric_df['layer'] = 'projection'
                    attn_metric_df = pd.DataFrame(attn_latency_eff_dict, index=[0])
                    attn_metric_df['layer'] = 'attention'
                    ffn_metric_df = pd.DataFrame(ffn_latency_eff_dict, index=[0])
                    ffn_metric_df['layer'] = 'ffn'
                    nonlinear_metric_df = pd.DataFrame(nonlinear_latency_eff_dict, index=[0])
                    nonlinear_metric_df['layer'] = 'nonlinear'


                    end_to_end_metric_df = pd.concat([proj_metric_df, attn_metric_df, ffn_metric_df, nonlinear_metric_df])
                    end_to_end_metric_df['arch'] = arch
                    end_to_end_metric_df['subarch'] = subarch
                    end_to_end_metric_df['arch_dim'] = arch_dim
                    end_to_end_metric_df['model'] = model

                    end_to_end_metric_df = end_to_end_metric_df.drop(columns=['flops', 'execution_time', 'power', 'energy'], errors='ignore')
                    end_to_end_breakdown = pd.concat([end_to_end_breakdown, end_to_end_metric_df], axis=0)

    end_to_end_breakdown.to_csv(output_path + 'end_to_end_latency_breakdown.csv', index=False)

    group_cols = [col for col in end_to_end_breakdown.columns if col not in ['latency', 'layer']]
    total_latency_df = end_to_end_breakdown.groupby(group_cols)['latency'].sum().reset_index()

    # Normalize each model separately against its own systolic mac 16x16 baseline
    end_to_end_breakdown['normalized_latency'] = 0.0  # Initialize column
    
    for model in end_to_end_breakdown['model'].unique():
        # Get baseline for this specific model
        baseline_df = total_latency_df[
            (total_latency_df['arch'] == 'mugi') &
            (total_latency_df['subarch'] == 'vlp') &
            (total_latency_df['arch_dim'] == '256x8') &
            (total_latency_df['model'] == model)
        ]
        
        if not baseline_df.empty:
            baseline_latency = baseline_df['latency'].values[0]
            # Normalize all entries for this model
            model_mask = end_to_end_breakdown['model'] == model
            end_to_end_breakdown.loc[model_mask, 'normalized_latency'] = (
                end_to_end_breakdown.loc[model_mask, 'latency'] / baseline_latency
            )

    end_to_end_breakdown.to_csv(output_path + 'end_to_end_latency_breakdown_norm.csv', index=False)

def figure(input_path: str, output_path: str):

    df = pd.read_csv(input_path + 'end_to_end_latency_breakdown_norm.csv')

    fig_width_pt = 250
    fig_width = fig_width_pt/72
    fig_height = fig_width/4.24

    font_size = 7

    fig, axes = plt.subplots(
        1, 4, figsize=(fig_width, fig_height), sharex=False, sharey=True
    )
    
    # Remove spacing between subplots to make them appear as one continuous graph
    plt.subplots_adjust(wspace=0)

    label_dict = {}

    for label in df.columns:
        if isinstance(df[label].iloc[0], str):
            label_dict[label] = list(df[label].unique())

    

    # Create extended architecture list for display (systolic subarchs as separate items)
    display_archs = []
    for arch in label_dict['arch']:
        if arch == 'simd':
            continue  # Skip simd architecture
        elif arch == 'systolic':
            display_archs.extend(['systolic', 'systolic_taylor', 'systolic_pwl'])
        else:
            display_archs.append(arch)
    
    bars_per_arch = {
        'mugi': 1,
        'carat': 1,
        'systolic': 1,
        'systolic_taylor': 1,
        'systolic_pwl': 1
    }

    bar_width = 0.5

    group_centers = np.arange(len(display_archs))


    colors_map = {
        'Mugi': "green",
        'Carat': "orange",
        'SA': "purple",
        'SA-F': "darkgray",
        'SD': "red",
        'SD-F': "dodgerblue"
    }

    # Colors for different layers in the stack
    layer_colors = {
        'projection': '#FF6B6B',  # Red
        'attention': "#53CC6D",   # Teal
        'ffn': '#45B7D1',         # Blue
        'nonlinear': '#FFA07A'    # Light Salmon
    }

    

    for idx, model in enumerate(label_dict['model']):
        ax = axes[idx]
        model_label = '7B' if model == 'llama_2_7b' else '13B' if model == 'llama_2_13b' else '70B' if model in 'llama_2_70b' else '70B GQA'
        ax.set_title(model_label, fontsize=font_size, pad = 2.5)

        

        x_pos = 0
        xticks = []
        xtick_labels = []

        for i, display_arch in enumerate(display_archs):
            # Determine actual arch and subarch for data filtering
            if display_arch.startswith('systolic'):
                arch = 'systolic'
                if display_arch == 'systolic':
                    subarch = 'mac'
                elif display_arch == 'systolic_taylor':
                    subarch = 'taylor'
                elif display_arch == 'systolic_pwl':
                    subarch = 'pwl'
            else:
                arch = display_arch
                subarch = ''
                
            n_bars = bars_per_arch.get(display_arch, 1)

            start_pos = group_centers[i] - (bar_width * n_bars) / 2
            bar_positions = start_pos + np.arange(n_bars) * bar_width

            np.empty((n_bars, len(label_dict['layer'])))
            
            if arch == 'mugi':
                filtered_df = df[
                    (df['model'] == model) &
                    (df['arch'] == arch) &
                    (df['subarch'] == 'vlp') &
                    (df['arch_dim'].isin(['256x8']))
                ]

            if arch == 'carat':
                filtered_df = df[
                    (df['model'] == model) &
                    (df['arch'] == arch) &
                    (df['arch_dim'].isin(['256x8']))
                ]
            
            if arch == 'systolic':
                filtered_df = df[
                    (df['model'] == model) &
                    (df['arch'] == arch) &
                    (df['subarch'] == subarch) &
                    (df['arch_dim'] == '16x16')
                ]

            if arch == 'tensor':
                filtered_df = df[
                    (df['model'] == model) &
                    (df['arch'] == arch) &
                    (df['arch_dim'] == '8x16x16')
                ]

            # Get data for each layer separately
            layer_data = {}
            for layer in ['projection', 'attention', 'ffn', 'nonlinear']:
                layer_filtered_df = filtered_df[filtered_df['layer'] == layer]
                layer_data[layer] = layer_filtered_df['normalized_latency'].tolist()

            # Create stacked bars for each architecture
            for j, pos in enumerate(bar_positions):
                bottom = 0
                for layer in ['projection', 'attention', 'ffn', 'nonlinear']:
                    if j < len(layer_data[layer]) and layer_data[layer]:
                        height = layer_data[layer][j] if j < len(layer_data[layer]) else 0
                        ax.bar(
                            pos,
                            height,
                            width=bar_width,
                            bottom=bottom,
                            label=layer if j == 0 and i == 0 else "",  # Only label once for legend
                            color=layer_colors[layer],
                            edgecolor='black',
                            linewidth=0.3
                        )
                        bottom += height

        # Calculate actual tick positions for each display architecture
        tick_positions = []
        display_arch_labels = []
        
        for i, display_arch in enumerate(display_archs):
            n_bars = bars_per_arch.get(display_arch, 1)
            start_pos = group_centers[i] - (bar_width * n_bars) / 2
            bar_positions = start_pos + np.arange(n_bars) * bar_width
            
            # For groups with multiple bars, position tick between bars
            # For single bars, position tick at the center of the bar
            if n_bars > 1:
                tick_pos = np.mean(bar_positions)
            else:
                tick_pos = bar_positions[0]
            tick_positions.append(tick_pos)
            
            # Create display labels
            if display_arch == 'systolic':
                display_arch_labels.append('S')
            elif display_arch == 'systolic_taylor':
                display_arch_labels.append('T')
            elif display_arch == 'systolic_pwl':
                display_arch_labels.append('P')
            elif display_arch == 'mugi':
                display_arch_labels.append('M')
            elif display_arch == 'carat':
                display_arch_labels.append('C')
            else:
                display_arch_labels.append(display_arch)
        
        ax.set_xticks(tick_positions)
        ax.set_xticklabels(display_arch_labels, fontsize=font_size-1)
        ax.set_xlim(-1, len(display_archs))
        
        # Add top x-axis with dimension labels
        ax2 = ax.twiny()
        
        # Calculate positions for merged dimension labels
        dim_tick_positions = []
        dim_tick_labels = []
        
        # Group architectures by dimension
        # VLP group: mugi, carat (256x8)
        vlp_indices = [i for i, arch in enumerate(display_archs) if arch in ['mugi', 'carat']]
        # Systolic group: systolic, systolic_taylor, systolic_pwl (16x16)
        systolic_indices = [i for i, arch in enumerate(display_archs) if arch.startswith('systolic')]
        
        # Calculate position for VLP group (256)
        if vlp_indices:
            vlp_positions = []
            for i in vlp_indices:
                n_bars = bars_per_arch.get(display_archs[i], 1)
                start_pos = group_centers[i] - (bar_width * n_bars) / 2
                bar_positions = start_pos + np.arange(n_bars) * bar_width
                if n_bars > 1:
                    vlp_positions.append(np.mean(bar_positions))
                else:
                    vlp_positions.append(bar_positions[0])
            
            # Center the label over the VLP group
            vlp_center = np.mean(vlp_positions)
            dim_tick_positions.append(vlp_center)
            dim_tick_labels.append('256')
        
        # Calculate position for Systolic group (16)
        if systolic_indices:
            systolic_positions = []
            for i in systolic_indices:
                n_bars = bars_per_arch.get(display_archs[i], 1)
                start_pos = group_centers[i] - (bar_width * n_bars) / 2
                bar_positions = start_pos + np.arange(n_bars) * bar_width
                if n_bars > 1:
                    systolic_positions.append(np.mean(bar_positions))
                else:
                    systolic_positions.append(bar_positions[0])
            
            # Center the label over the Systolic group
            systolic_center = np.mean(systolic_positions)
            dim_tick_positions.append(systolic_center)
            dim_tick_labels.append('16')
        
        ax2.set_xticks(dim_tick_positions)
        ax2.set_xticklabels(dim_tick_labels, fontsize=font_size-1)
        ax2.set_xlim(ax.get_xlim())
        
        # Make tick marks thinner
        ax.tick_params(axis='both', width=0.3, length=2)
        ax2.tick_params(axis='x', width=0.3, length=2)
        
        # Make both xticks closer to the graph
        ax.tick_params(axis='x', pad=0.5)
        ax2.tick_params(axis='x', pad=0.25)
        
        # Set y-axis tick font size to match x-axis
        ax.tick_params(axis='y', labelsize=font_size)
        
        # Set y-axis to show integers 0, 1, 2 only
        ax.set_yticks([0, 1, 2])
        ax.set_yticklabels(['0', '1', '2'])
        ax.set_ylim(0, 2.2)  # Set y-limit to accommodate the 2.0 max
        
        # Remove all spines except outer ones to create unified appearance
        if idx > 0:
            # Remove y-axis labels and ticks for interior subplots
            ax.tick_params(axis='y', left=False, labelleft=False)
            # Add light dashed grid lines (same as leftmost subplot)
            ax.grid(True, axis='y', linestyle='--', linewidth=0.3, alpha=0.5, color='gray')
            # Remove left spine for interior subplots
            ax.spines['left'].set_visible(False)
        else:
            # Ensure y-axis is visible on the leftmost subplot with values
            ax.tick_params(axis='y', left=True, labelleft=True, labelsize=font_size)
            # Explicitly enable y-axis labels
            ax.yaxis.set_tick_params(labelleft=True)
            # Add light dashed grid lines
            ax.grid(True, axis='y', linestyle='--', linewidth=0.3, alpha=0.5, color='gray')
            # Make leftmost spine thinner
            ax.spines['left'].set_linewidth(0.3)
        
        # Remove right spine for all but the rightmost subplot
        if idx < len(label_dict['model']) - 1:
            ax.spines['right'].set_visible(False)
        else:
            # Make rightmost spine thinner
            ax.spines['right'].set_linewidth(0.3)
        
        # Make top and bottom spines thinner
        ax.spines['top'].set_linewidth(0.3)
        ax.spines['bottom'].set_linewidth(0.3)
        
        # Remove top spine from twin axis except for outer ones
        if idx == 0:
            ax2.spines['top'].set_linewidth(0.3)
        else:
            ax2.spines['top'].set_visible(False)
            
        # Remove other spines from twin axis
        ax2.spines['bottom'].set_visible(False)
        ax2.spines['left'].set_visible(False)
        ax2.spines['right'].set_visible(False)

    # Remove xlabel from all subplots (was previously added)
    # for ax in axes:
    #     ax.set_xlabel("Architecture", fontsize=font_size)
    
    # Add legend for layers at the top in one row
    handles = [plt.Rectangle((0,0),1,1, color=layer_colors[layer]) for layer in ['projection', 'attention', 'ffn', 'nonlinear']]
    labels = ['Projection', 'Attention', 'FFN', 'Nonlinear']
    fig.legend(handles, labels, loc='upper center', bbox_to_anchor=(0.525, 1.4), ncol=4, fontsize=font_size,
               columnspacing=1, handlelength=.75, handleheight=0.5, handletextpad=1)
    plt.subplots_adjust(left=0.08, right=0.98, top=0.82, bottom=0.25)

    plt.savefig(output_path + 'end_to_end_latency_breakdown.png', dpi=1200, bbox_inches='tight')
    plt.savefig(output_path + 'end_to_end_latency_breakdown.pdf', dpi=1200, bbox_inches='tight')


from collections import OrderedDict

from archx.utils import get_prod


def gemm16(architecture_dict: OrderedDict, workload_dict: OrderedDict=None):
    workload_dict = workload_dict['gemm16']
    performance_dict = gemm(architecture_dict, workload_dict)
    return performance_dict


def gemm32(architecture_dict: OrderedDict, workload_dict: OrderedDict=None):
    workload_dict = workload_dict['gemm32']
    performance_dict = gemm(architecture_dict, workload_dict)
    return performance_dict


def gemm(architecture_dict: OrderedDict, workload_dict: OrderedDict=None):
    # this is a very simple performance model for matrix multiplication
    performance_dict = OrderedDict()

    mac_array_size = get_prod(architecture_dict['multiplier']['instance'])
    mac_total = workload_dict['configuration']['m'] * workload_dict['configuration']['k'] * workload_dict['configuration']['n']
    # number of calls to mac_array
    mac_array_total = mac_total / mac_array_size

    # number of memory accesses
    # no data reuse is considered
    word_size = architecture_dict['multiplier']['query']['bitwidth']
    sram_rd_num = word_size * 2 * mac_total / architecture_dict['sram']['query']['width']
    sram_wr_num = sram_rd_num / 2

    mac_array_dict = OrderedDict({'count': mac_array_total})
    sram_rd_dict = OrderedDict({'count': sram_rd_num})
    sram_wr_dict = OrderedDict({'count': sram_wr_num})
    
    performance_dict['subevent'] = OrderedDict({'mac_array': mac_array_dict, 
                                                'sram_rd': sram_rd_dict, 
                                                'sram_wr': sram_wr_dict})

    return performance_dict


def mac_array(architecture_dict: OrderedDict, workload_dict: OrderedDict=None):
    # this is an array of mac units
    performance_dict = OrderedDict()

    # 1 cycle per mac
    # as mac units are parallel, 1 cycle per mac_array
    performance_dict['cycle_count'] = OrderedDict({'value': 1., 'unit': 'cycles'})
    performance_dict['runtime'] = OrderedDict({'value': 1 / 1000 / architecture_dict['multiplier']['query']['frequency'] , 'unit': 'ms'})

    # by default, architecture modules have
    # 'count' is 1
    # 'aggregation' is 'parallel'

    mac_array_size = get_prod(architecture_dict['multiplier']['instance'])
    multiplier_dict = OrderedDict({'count': mac_array_size})
    adder_dict = OrderedDict({'count': mac_array_size})
    
    performance_dict['subevent'] = OrderedDict({'multiplier' : multiplier_dict, 
                                                'adder' : adder_dict})

    return performance_dict


def sram_rd(architecture_dict: OrderedDict, workload_dict: OrderedDict=None):
    performance_dict = OrderedDict()
    performance_dict['cycle_count'] = OrderedDict({'value': 1., 'unit': 'cycles'})
    performance_dict['runtime'] = OrderedDict({'value': 1 / 1000 / architecture_dict['sram']['query']['frequency'] , 'unit': 'ms'})
    sram_dict = OrderedDict({'operation': OrderedDict({'dynamic_energy' : 'read'})})
    performance_dict['subevent'] = OrderedDict({'sram' : sram_dict})
    return performance_dict


def sram_wr(architecture_dict: OrderedDict, workload_dict: OrderedDict=None):
    performance_dict = OrderedDict()
    performance_dict['cycle_count'] = OrderedDict({'value': 1., 'unit': 'cycles'})
    performance_dict['runtime'] = OrderedDict({'value': 1 / 1000 / architecture_dict['sram']['query']['frequency'] , 'unit': 'ms'})
    sram_dict = OrderedDict({'operation': OrderedDict({'dynamic_energy' : 'write'})})
    performance_dict['subevent'] = OrderedDict({'sram' : sram_dict})
    return performance_dict




from zoo.llm.results.query.utils import query_area, query_performance_gemm_metrics, query_performance_nonlinear_metrics, load_yaml, compute_throughput_efficiancy
import pandas as pd
from collections import OrderedDict
import os

def query(input_path, output_path):
    network_list = ['single_node', 'multi_node_4x4', 'multi_node_8x8']
    tensor_network_list = ['single_node', 'multi_node_2x1', 'multi_node_2x2']

    vlp_list = ['mugi', 'carat']
    vlp_arch_dim_list = ['64x8', '128x8', '256x8']

    baseline_list = ['systolic', 'simd']
    baseline_arch_dim_list = ['8x8', '16x16', '64x64']
    baseline_subarch_list = ['mac', 'figna']

    throughput_module = OrderedDict({
        'mugi': OrderedDict({'gemm': 'and_gate', 'nonlinear': 'magnitude_register'}),
        'carat': OrderedDict({'gemm': 'and_gate', 'nonlinear': 'accumulator_vector'}),
        'systolic': OrderedDict({'gemm': 'multiplier', 'nonlinear': 'accumulator_vector'}),
        'simd': OrderedDict({'gemm': 'multiplier', 'nonlinear': 'accumulator_vector'}),
        'tensor': OrderedDict({'gemm': 'multiplier', 'nonlinear': 'accumulator_vector'}),
    })

    model = 'llama_2_70b_GQA'
    max_seq_len = 'max_seq_len_4096'
    batch_size = 'batch_size_8'

    kv_heads = 'kv_heads_8'

    comprehensive_table_df = pd.DataFrame()

    for arch in (vlp_list + baseline_list) + ['tensor']:
        for network in network_list if arch != 'tensor' else tensor_network_list:
            for subarch in (baseline_subarch_list if arch in baseline_list else ['vlp'] if arch == 'mugi' else ['']):
                for arch_dim in (vlp_arch_dim_list if arch in vlp_list else baseline_arch_dim_list if arch in baseline_list else ['8x16x16'] if arch == 'tensor' else ['']):
                    if arch_dim in ['64x64', '128x8'] and network != 'single_node':
                        continue
                    if arch_dim in ['8x8', '64x8'] and network == 'single_node':
                        continue
                    if arch_dim in ['256x8', '16x16'] and network == 'multi_node_8x8':
                        continue
                    if arch_dim in ['64x8', '8x8'] and network == 'multi_node_4x4':
                        continue

                    gemm_module = throughput_module[arch]['gemm']
                    nonlinear_module = throughput_module[arch]['nonlinear']
                    
                    termination_path = 'full_termination' if arch == 'mugi' else ''
                    run_path = os.path.normpath(f'{input_path}{arch}/{network}/{subarch}/{arch_dim}/{model}/{max_seq_len}/{batch_size}/kv_heads_8/{termination_path}/')
                    yaml_dict = load_yaml(run_path)

                    event_graph = yaml_dict['event_graph']
                    metric_dict = yaml_dict['metric_dict']

                    gemm_performance_metrics_dict = query_performance_gemm_metrics(event_graph=event_graph, metric_dict=metric_dict, workload=model, event='gemm', module=gemm_module)
                    nonlinear_performance_metrics_dict = query_performance_nonlinear_metrics(event_graph=event_graph, metric_dict=metric_dict, workload=model, event='nonlinear', module=nonlinear_module)

                    assert gemm_performance_metrics_dict['power'] == nonlinear_performance_metrics_dict['power'], "Power mismatch between gemm and nonlinear modules"

                    performance_metrics_dict = OrderedDict({
                        'flops': gemm_performance_metrics_dict['flops'] + nonlinear_performance_metrics_dict['flops'],
                        'execution_time': gemm_performance_metrics_dict['execution_time'] + nonlinear_performance_metrics_dict['execution_time'],
                        'energy': gemm_performance_metrics_dict['energy'] + nonlinear_performance_metrics_dict['energy'],
                        'power': gemm_performance_metrics_dict['power']
                    })

                    throughput_eff_dict = compute_throughput_efficiancy(performance_metrics_dict=performance_metrics_dict)
                    area = query_area(event_graph=event_graph, metric_dict=metric_dict, tag='onchip')

                    comprehensive_table_dict = OrderedDict({
                        'arch': arch,
                        'subarch': subarch,
                        'network': network,
                        'arch_dim': arch_dim,
                        'throughput': throughput_eff_dict['throughput'],
                        'energy_efficiency': throughput_eff_dict['energy_efficiency'] * (10**3), # GFLOPS/s/mJ
                        'power_efficiency': throughput_eff_dict['power_efficiency'],
                        'area': area,
                        'flops': performance_metrics_dict['flops'],
                        'execution_time': performance_metrics_dict['execution_time'],
                        'energy': performance_metrics_dict['energy'],
                        'power': performance_metrics_dict['power'],
                        'gemm_flops': gemm_performance_metrics_dict['flops'],
                        'gemm_execution_time': gemm_performance_metrics_dict['execution_time'],
                        'nonlinear_flops': nonlinear_performance_metrics_dict['flops'],
                        'nonlinear_execution_time': nonlinear_performance_metrics_dict['execution_time'],
                    })

                    comprehensive_table_df = pd.concat([comprehensive_table_df, pd.DataFrame(comprehensive_table_dict, index=[0])])
                    
    comprehensive_table_df.to_csv(output_path + 'comprehensive_table.csv', index=False)

def table(input_path: str, output_path: str):
    df = pd.read_csv(input_path + 'comprehensive_table.csv')
    
    vlp_single_node_list = []
    baseline_single_node_list = []
    baseline_sa_list = []
    vlp_multi_node_list = []
    sa_multi_node_list = []
    sa_f_multi_node_list = []
    sd_multi_node_list = []
    sd_f_multi_node_list = []
    tensor_single_node_list = []
    tensor_multi_node_list = []
    for index, row in df.iterrows():
        arch_label = row['arch'].capitalize()
        dim_label = row['arch_dim'].split('_')[0]
        subarch_label = row['subarch'] if pd.notna(row['subarch']) else ''
        network = row['network']

        if arch_label == 'Systolic':
            arch_label = 'SA'
        elif arch_label == 'Simd':
            arch_label = 'SD'
        if arch_label == 'Mugi':
            arch_label = '\\name'
        if arch_label == 'tensor':
            arch_label = 'Tensor'

        subarch_label = '-F' if subarch_label == 'figna' else ''

        design = arch_label + subarch_label + ' (' + dim_label.split('x')[0] + ')'
        throughput = row['throughput']
        area = row['area']
        energy_efficiency = row['energy_efficiency']
        power_efficiency = row['power_efficiency']
        if 'multi_node' in network:
            network_label = network.split('_')[2].split('x')
            network_label = ' & ' + network_label[0] + ' x ' + network_label[1]
        else:
            network_label = ''

        string = f"{network_label} & {design} & {throughput:.2f} & {area:.2f} & {energy_efficiency:.2f} & {power_efficiency:.2f} \\\\"
        if arch_label in ['\\name', 'Carat'] and network == 'single_node':
            vlp_single_node_list.append(string)
        elif arch_label in ['SA', 'SD'] and network == 'single_node' and dim_label in ['8x8', '16x16']:
            baseline_single_node_list.append(string)
        elif arch_label in ['SA', 'SD'] and network == 'single_node' and dim_label in ['64x64']:
            baseline_sa_list.append(string)
        elif arch_label in ['\\name', 'Carat'] and network != 'single_node':
            vlp_multi_node_list.append(string)
        elif arch_label in ['SA'] and network != 'single_node':
            if subarch_label == '-F':
                sa_f_multi_node_list.append(string)
            else:
                sa_multi_node_list.append(string)
        elif arch_label in ['SD'] and network != 'single_node':
            if subarch_label == '-F':
                sd_f_multi_node_list.append(string)
            else:
                sd_multi_node_list.append(string)
        elif arch_label == 'Tensor' and network == 'single_node':
            tensor_single_node_list.append(string)
        elif arch_label == 'Tensor' and network != 'single_node':
            tensor_multi_node_list.append(string)
        
    with open(output_path + 'comprehensive_table.txt', 'w') as f:
        f.write(' & Arch & Throughput & Area & Energy Efficiency & Power Efficiency \\\\\n')
        f.write(' &      & GFLOP/s & mm^2 & GFLOP/s/mJ & GFLOP/s/W \\\\\n')
    with open(output_path + 'comprehensive_table.txt', 'a') as f:
        for string in vlp_single_node_list:
            f.write(string + '\n')
        for string in baseline_single_node_list:
            f.write(string + '\n')
        for string in baseline_sa_list:
            f.write(string + '\n')
        for string in tensor_single_node_list:
            f.write(string + '\n')
        for string in vlp_multi_node_list:
            f.write(string + '\n')
        for string in sa_multi_node_list:
            f.write(string + '\n')
        for string in sa_f_multi_node_list:
            f.write(string + '\n')
        for string in sd_multi_node_list:
            f.write(string + '\n')
        for string in sd_f_multi_node_list:
            f.write(string + '\n')
        for string in tensor_multi_node_list:
            f.write(string + '\n')


from archx.utils import read_yaml, write_yaml
from collections import OrderedDict
import os, shutil, copy, itertools
from loguru import logger


def clean_dir(base_path: str):
    base_path = base_path if base_path.endswith('/') else base_path + '/'
    generated_path = f'{base_path}generated/'
    shutil.rmtree(generated_path, ignore_errors=True)

def generate_architecture(base_path: str):
    """
    Generate architecture yamls from template
    """

    base_path = base_path if base_path.endswith('/') else base_path + '/'
    template_path = base_path + 'template/'
    generated_path = base_path + 'generated/'

    architecture_file = template_path + 'architecture.yaml'
    configuration_file = template_path + 'configuration.yaml'

    architecture_dict = read_yaml(architecture_file)
    configuration_dict = read_yaml(configuration_file)

    network_config_dict = configuration_dict.get('network')
    tag_config_dict = configuration_dict.get('tag')
    array_config_dict = configuration_dict.get('array')

    subarchitecture_config_dict = configuration_dict.get('subarchitecture', [None])
    
    # Pass through each network, subarchitecture, and array configuration
    # - Network: single/multinode, node dimensions, ex. 4x4, 8x8
    # - subarchitecture: ex. mac, figna, etc.
    # - array: array dimensions, ex. mugi 128x8, 256x8, systolic 8x8, 16x16, etc.

    for network_key, network_value in network_config_dict.items():
        for network_dimension in network_value['dimension']:
            for subarchitecture in subarchitecture_config_dict:
                for array_dim, array_value in array_config_dict.items():
                    if network_dimension == []:
                        network_path = f'{network_key}/'
                    else:
                        network_path = f'{network_key}_{network_dimension[0]}x{network_dimension[1]}/'

                    # can be dictionary of instance, width, depth labels or path to yaml containing those labels
                    if 'path' in array_value:
                        array_value = read_yaml(array_value['path'])
                        
                    generated_architecture_dict = copy.deepcopy(architecture_dict)

                    if 'path' in network_value:
                        network_architecture_dict = read_yaml(network_value['path'])
                        generated_architecture_dict['architecture']['module'].update(network_architecture_dict['architecture']['module'])

                    if subarchitecture:
                        subarchitecture_dict = read_yaml(subarchitecture_config_dict[subarchitecture]['path'])
                        generated_architecture_dict['architecture']['module'].update(subarchitecture_dict['architecture']['module'])
                        subarchitecture_path = f'{subarchitecture}/'
                    else:
                        subarchitecture_path = '/'

                    generated_architecture_component_dict = OrderedDict({'architecture': {'module': {}}})
                    for component_key, component_value in generated_architecture_dict['architecture']['module'].items():
                        if 'path' in component_value:
                            component_value = read_yaml(component_value['path'])['architecture']['module']
                            generated_architecture_component_dict['architecture']['module'].update(component_value)
                        else:
                            generated_architecture_component_dict['architecture']['module'][component_key] = component_value

                    for component_key, component_value in generated_architecture_component_dict['architecture']['module'].items():
                        if isinstance(component_value.get('tag'), str):
                            component_value['tag'] = copy.deepcopy(tag_config_dict.get(component_value['tag'], None))

                        if isinstance(component_value.get('instance'), str):
                            component_value['instance'] = network_dimension + array_value['instance'].get(component_value['instance'], None)

                        if isinstance(component_value['query'].get('width'), str):
                            component_value['query']['width'] = array_value['width'].get(component_value['query']['width'], None)
                        if isinstance(component_value['query'].get('depth'), str):
                            component_value['query']['depth'] = array_value['depth'].get(component_value['query']['depth'], None)
                    generated_architecture_component_dict['architecture']['attribute'] = generated_architecture_dict['architecture']['attribute']

                    output_path = f'{generated_path}{subarchitecture_path}{network_path}{array_dim}/architecture.yaml'
                    write_yaml(file=output_path, content=generated_architecture_component_dict)

def generate_event(base_path: str):
    """
    Generate architecture yamls from template
    """

    base_path = base_path if base_path.endswith('/') else base_path + '/'
    template_path = base_path + 'template/'
    generated_path = base_path + 'generated/'

    configuration_dict = read_yaml(template_path + 'configuration.event.yaml')
    network_config_dict = configuration_dict.get('network')
    hardware_config_dict = configuration_dict.get('hardware')
    default_config_dict = configuration_dict.get('default')
    memory_config_dict = default_config_dict.get('memory')

    for network_key, network_value in network_config_dict.items():
        for subarchitecture_key, subarchitecture_value in hardware_config_dict.items():
            generated_event_dict = OrderedDict({'event': {}})

            for default_event_value in default_config_dict.values():
                if 'path' in default_event_value:
                    default_event_dict = read_yaml(default_event_value['path'])
                generated_event_dict['event'].update(default_event_dict['event'])
                
                if network_value:
                    if 'path' in network_value:
                        network_event_dict = read_yaml(network_value['path'])
                    generated_event_dict['event'].update(network_event_dict['event'])

                for hardware_key, hardware_value in subarchitecture_value.items():
                    if 'path' in hardware_value:
                        hardware_event_dict = read_yaml(hardware_value['path'])
                    generated_event_dict['event'].update(hardware_event_dict['event'])

                for hardware_key, hardware_value in subarchitecture_value.items():
                    if 'path' in hardware_value:
                        hardware_event_dict = read_yaml(hardware_value['path'])
                    for event_key, event_value in generated_event_dict['event'].items():
                        if not isinstance(event_value['subevent'], list) and event_value['subevent'] == hardware_key:
                            hardware_subevent_list = list(hardware_event_dict['event'].keys())
                            if 'path' in memory_config_dict:
                                memory_config_dict = read_yaml(memory_config_dict['path'])
                            memory_subevent_list = list(memory_config_dict['event'].keys())
                            network_subevent_list = list(network_event_dict['event'].keys()) if 'network_event_dict' in locals() else []
                            event_value['subevent'] = hardware_subevent_list + memory_subevent_list + network_subevent_list

                generated_event_path = f'{generated_path}{network_key}/{subarchitecture_key}/event.yaml'
                write_yaml(generated_event_path, generated_event_dict)

def generate_workload(base_path: str):
    """
    Generate architecture yamls from template
    """

    base_path = base_path if base_path.endswith('/') else base_path + '/'
    template_path = base_path + 'template/'
    generated_path = base_path + 'generated/'

    workload_dict = read_yaml(template_path + 'workload.yaml')
    configuration_dict = read_yaml(template_path + 'configuration.yaml')
    config_dict = configuration_dict.get('configuration')
    architecture_dict = configuration_dict.get('architecture', {'workload': None})

    for workload_key, workload_value in workload_dict['workload'].items():
        for architecture_key, architecture_value in architecture_dict.items():
            if architecture_key == 'workload':
                architecture_key = ''
            filtered_config_dict = copy.deepcopy(config_dict)
            shared_keys = set(workload_value['configuration'].keys()) & set(filtered_config_dict.keys())
            for shared_key in shared_keys:
                del filtered_config_dict[shared_key]
            config_combinations = [dict(zip(filtered_config_dict.keys(), combination)) for combination in itertools.product(*filtered_config_dict.values())]
            for combination in config_combinations:
                generated_workload_dict = OrderedDict({'workload': {workload_key: {'configuration': {}}}})
                generated_workload_dict['workload'][workload_key]['configuration'].update(workload_value['configuration'])
                if architecture_value:
                    generated_workload_dict['workload'][workload_key]['configuration'].update(architecture_value)
                generated_workload_dict['workload'][workload_key]['configuration'].update(combination)

                workload_config_path = ''
                for key, value in combination.items():
                    if len(filtered_config_dict[key]) > 1:
                        workload_config_path += '/' + key + '_' + str(value)

                workload_path = os.path.normpath(f'{generated_path}{workload_key}/{architecture_key}/{workload_config_path}/workload.yaml')
                write_yaml(workload_path, generated_workload_dict)

def main():
    logger.remove()
    base_dir = 'zoo/llm/'
    architecture_list = read_yaml(base_dir + 'scripts/architecture_configuration.yaml')
    for arch in architecture_list['architecture']:
        base_path = os.path.join(base_dir, arch)
        if os.path.isdir(base_path):
            architecture_path = os.path.join(base_path, 'architecture/')
            event_path = os.path.join(base_path, 'event/')
            workload_path = os.path.join(base_path, 'workload/')

            clean_dir(architecture_path)
            clean_dir(event_path)
            clean_dir(workload_path)

            generate_architecture(architecture_path)
            generate_event(event_path)
            generate_workload(workload_path)

if __name__ == "__main__":
    main()


from loguru import logger

class Metric:
    def __init__(self):
        self.metrics = {'metric': {}}

    def add_metric(self, name: str, unit: str, aggregation: str):
        assert isinstance(name, str), "'name' parameter must be of type 'str'."
        assert isinstance(unit, str), "'unit' parameter must be of type 'str'."
        assert isinstance(aggregation, str), "'aggregation' parameter must be of type 'str'."

        self.metrics['metric'][name] = {
            "unit": unit,
            "aggregation": aggregation
        }

        logger.info(f"Added metric '{name}'.")
        logger.debug(f"\tUnit: {unit}")
        logger.debug(f"\tAggregation: {aggregation}")

    def to_yaml(self):
        return self.metrics


from collections import OrderedDict
from loguru import logger

from archx.utils import read_yaml, get_path, write_yaml


key_path = 'path'
key_configuration = 'configuration'
key_workload = 'workload'


def create_workload_dict(workload_file: str) -> OrderedDict:
    """
    Create a workload dict, path are loaded recursively.
    """
    # read in workload yaml
    workload_file_full_path = get_path(workload_file)
    workload_dict = read_yaml(workload_file_full_path)[key_workload]

    for workload in workload_dict:
        if key_path not in workload_dict[workload]:
            assert key_configuration in workload_dict[workload], logger.info(f'No configuration in workload <{workload}>.')
        else:
            # load workload from path
            new_workload_file = get_path(workload_dict[workload][key_path])
            workload_dict[workload] = create_workload_dict(new_workload_file)[workload]
    
    logger.success(f'Create workload dictionary from <{workload_file_full_path}>.')

    return workload_dict


def save_workload_dict(workload_dict: OrderedDict, save_path: str) -> None:
    """
    Save workload to a checkpoint
    """
    save_path = get_path(save_path, check_exist=False)
    workload_dict_ckpt = OrderedDict({key_workload: workload_dict})
    write_yaml(save_path, workload_dict_ckpt)
    logger.success(f'Save workload dictionary to <{save_path}>.')


def load_workload_dict(ckpt_path: str) -> OrderedDict:
    """
    Load workload from a checkpoint
    """
    full_path = get_path(ckpt_path)
    workload_dict = read_yaml(full_path)[key_workload]
    logger.success(f'Load workload dictionary from <{full_path}>.')
    return workload_dict

